@@ -236,6 +236,7 @@ static int bam_get_line(bam_file_t *b, unsigned char **str, size_t *len) {
 	    if (*from != '\n') {
 		*to++ = *from++;
 	    } else {
+		if (to[-1] == '\r') *--to = 0; // handle \r\n too
 		b->uncomp_p = from;
 		used_l = to-buf;
 		b->uncomp_p++;
@@ -253,7 +254,9 @@ static int bam_get_line(bam_file_t *b, unsigned char **str, size_t *len) {
 
 	if (used_l >= alloc_l) {
 	    alloc_l = alloc_l ? alloc_l * 2 : 1024;
-	    if (NULL == (buf = realloc(buf, alloc_l)))
+	    // +8 to cope with the 64-bit copy function in the
+	    // COPY_CPF_TO_CPTM macro.
+	    if (NULL == (buf = realloc(buf, alloc_l+8)))
 		return -1;
 	}
     }
@@ -1025,19 +1028,9 @@ static int bam_uncompress_input(bam_file_t *b) {
 #endif
 
 
-#define STRTOL(v,rv,b) ({			\
-	    int n = 0;				\
-	    while (*v > '\t')			\
-		n=n*10+*v++-'0';		\
-	    *rv=v;				\
-	    n;					\
-	})
-
-#undef STRTOL
-
 /* Custom strtol for aux tags, always base 10 */
-static long inline STRTOL(const char *v, const char **rv, int b) {
-    long n = 0;
+static int64_t inline STRTOL64(const char *v, const char **rv, int b) {
+    int64_t n = 0;
     int neg = 1;
     switch(*v) {
     case '-':
@@ -1365,6 +1358,7 @@ static int sam_next_seq(bam_file_t *b, bam_seq_t **bsp) {
 	*cpt++ = key[1];
 
 	switch(key[3]) {
+	    int64_t n;
 	case 'A':
 	    *cpt++ = 'A';
 	    *cpt++ = *value;
@@ -1372,7 +1366,7 @@ static int sam_next_seq(bam_file_t *b, bam_seq_t **bsp) {
 
 	case 'i':
 	    //n = atoi((char *)value);
-	    n = STRTOL((char *)value, (const char **)&value, 10);
+	    n = STRTOL64((char *)value, (const char **)&value, 10);
 	    if (n >= 0) {
 		if (n < 256) {
 		    *cpt++ = 'C';
@@ -1796,18 +1790,23 @@ char *bam_aux_Z(const uint8_t *dat) {
  * An iterator on bam aux fields. NB: This code is not reentrant or multi-
  * thread capable. The values returned are valid until the next call to
  * this function.
- * key:  points to an array of 2 characters (eg "RG", "NM")
- * type: points to an address of 1 character (eg 'Z', 'i')
+ * key:  points to an array of 3 characters (eg "RGi", "NMC")
+ * type: points to an address of 1 character (eg 'Z', 'i') for the SAM type
  * val:  points to an address of a bam_aux_t union.
  *
+ * The first two bytes of key are the real key and the next byte is the
+ * BAM type field. Note that this may differ to the returned SAM type
+ * field. For example key[2] == 'S' for unsigned short while *type is
+ * set to 'i'.
+ *
  * Pass in *iter_handle as NULL to initialise the search and then
  * pass in the modified value on each subsequent call to continue the search.
  *
  * Returns 0 if the next value is valid, setting key, type and val.
  *        -1 when no more found.
  */
-int bam_aux_iter(bam_seq_t *b, char **iter_handle,
-		 char *key, char *type, bam_aux_t *val) {
+int bam_aux_iter_full(bam_seq_t *b, char **iter_handle,
+		      char *key, char *type, bam_aux_t *val) {
     char *s;
 
     if (!iter_handle || !*iter_handle) {
@@ -1822,6 +1821,7 @@ int bam_aux_iter(bam_seq_t *b, char **iter_handle,
 
     key[0] = s[0];
     key[1] = s[1];
+    key[2] = s[2];
     
     switch (s[2]) {
     case 'A':
@@ -1948,6 +1948,24 @@ int bam_aux_iter(bam_seq_t *b, char **iter_handle,
     return 0;
 }
 
+/*
+ * As above, but only 2 characters of the key are returned so the
+ * original BAM type is not visible.
+ *
+ * Note this can cause ambiguities if you wish to distinguish between
+ * -1 billion and +3 billion.
+ */
+int bam_aux_iter(bam_seq_t *b, char **iter_handle,
+		 char *key, char *type, bam_aux_t *val) {
+    char k3[3];
+    int r = bam_aux_iter_full(b, iter_handle, k3, type, val);
+
+    key[0] = k3[0];
+    key[1] = k3[1];
+
+    return r;
+}
+
 static int reg2bin(int start, int end) {
     if (end>start) end--;
     if ((start>>14) == (end>>14)) return ((1<<15)-1)/7 + (start>>14);
@@ -2060,16 +2078,18 @@ int bam_construct_seq(bam_seq_t **b, size_t extra_len,
     cp += ncigar*4;
 
     /* Bin */
-    if (0 == end) { /* Calculate end from pos and cigar */
-	end = pos;
-	for (i = 0; i < ncigar; i++) {
-	    int op = cigar[i] & BAM_CIGAR_MASK;
-	    if (op == BAM_CMATCH || op == BAM_CDEL || op == BAM_CREF_SKIP)
-		end += cigar[i] >> BAM_CIGAR_SHIFT;
+    if (!((*b)->flag & BAM_CIGAR32)) {
+	if (0 == end) { /* Calculate end from pos and cigar */
+	    end = pos;
+	    for (i = 0; i < ncigar; i++) {
+		if (BAM_CONSUME_REF(cigar[i] & BAM_CIGAR_MASK))
+		    end += cigar[i] >> BAM_CIGAR_SHIFT;
+	    }
 	}
-    }
 
-    bam_set_bin(*b, reg2bin(pos-1,end-1));
+	// range is [beg,end) and zero based.
+	bam_set_bin(*b, reg2bin(pos-1,end));
+    }
 
     /* Seq */
     for (i = 0; i < len-1; i += 2) {
@@ -2366,6 +2386,8 @@ int bam_aux_add_from_sam(bam_seq_t **bsp, char *sam) {
 	*cpt++ = key[1];
 
 	switch(key[3]) {
+	    int64_t n;
+
 	case 'A':
 	    *cpt++ = 'A';
 	    *cpt++ = *value;
@@ -2373,7 +2395,7 @@ int bam_aux_add_from_sam(bam_seq_t **bsp, char *sam) {
 
 	case 'i':
 	    //n = atoi((char *)value);
-	    n = STRTOL((char *)value, (const char **)&value, 10);
+	    n = STRTOL64((char *)value, (const char **)&value, 10);
 	    if (n >= 0) {
 		if (n < 256) {
 		    *cpt++ = 'C';
@@ -2595,7 +2617,7 @@ bam_seq_t *bam_dup(bam_seq_t *b) {
 }
 
 
-static unsigned char *append_int(unsigned char *cp, int32_t i) {
+unsigned char *append_int(unsigned char *cp, int32_t i) {
     int32_t j;
 
     if (i < 0) {
@@ -2651,7 +2673,7 @@ static unsigned char *append_int(unsigned char *cp, int32_t i) {
  * Unsigned version of above.
  * Only differs when the int has the top bit set (~2.15 billion and above).
  */
-static unsigned char *append_uint(unsigned char *cp, uint32_t i) {
+unsigned char *append_uint(unsigned char *cp, uint32_t i) {
     uint32_t j;
 
     if (i == 0) {
@@ -2879,8 +2901,16 @@ static int bgzf_flush(bam_file_t *bf) {
  * Returns 0 on success
  *        -1 on failure
  */
+#if defined(__GNUC__) && !(defined(__clang__) || defined(__ICC))
+// On later gcc releases the ALLOW_UAC code causes the vectorizor to
+// use SIMD instructions on unaligned memory access.  This is due to
+// our own abuse of char to int aliasing, but doing things the legal
+// way is still slower overall (5-14% depending on system and
+// compiler).
+__attribute__((optimize("no-tree-vectorize")))
+#endif
 int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
-    char *auxh, aux_key[2], type;
+    char *auxh, aux_key[3], type;
     bam_aux_t val;
 
     /*
@@ -3151,7 +3181,7 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
 
 	/* Auxiliary tags */
 	auxh = NULL;
-	while (0 == bam_aux_iter(b, &auxh, aux_key, &type, &val)) {
+	while (0 == bam_aux_iter_full(b, &auxh, aux_key, &type, &val)) {
 	    if (end - fp->uncomp_p < 20) BF_FLUSH();
 	    *fp->uncomp_p++ = '\t';
 	    *fp->uncomp_p++ = aux_key[0];
@@ -3159,13 +3189,13 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
 	    *fp->uncomp_p++ = ':';
 	    *fp->uncomp_p++ = type;
 	    *fp->uncomp_p++ = ':';
-	    switch(type) {
+	    switch(aux_key[2]) {
 	    case 'A':
 		*fp->uncomp_p++ = val.i;
 		break;
 
 	    case 'C':
-		fp->uncomp_p = append_int(fp->uncomp_p, (uint8_t)val.i);
+		fp->uncomp_p = append_uint(fp->uncomp_p, (uint8_t)val.i);
 		break;
 
 	    case 'c':
@@ -3173,7 +3203,7 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
 		break;
 
 	    case 'S':
-		fp->uncomp_p = append_int(fp->uncomp_p, (uint16_t)val.i);
+		fp->uncomp_p = append_uint(fp->uncomp_p, (uint16_t)val.i);
 		break;
 
 	    case 's':
@@ -3181,7 +3211,7 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
 		break;
 
 	    case 'I':
-		fp->uncomp_p = append_int(fp->uncomp_p, (uint32_t)val.i);
+		fp->uncomp_p = append_uint(fp->uncomp_p, (uint32_t)val.i);
 		break;
 
 	    case 'i':
