@@ -205,10 +205,10 @@ typedef struct {
 #define bam_mate_ref(b)  ((b)->mate_ref)
 #define bam_mate_pos(b)  ((b)->mate_pos)
 #define bam_ins_size(b)  ((b)->ins_size)
-#define bam_cigar_len(b) ((b)->cigar_len)
+#define bam_cigar_len(b) (((b)->flag & BAM_CIGAR32 ? ((b)->bin<<16) : 0) + (b)->cigar_len)
 #define bam_name_len(b)  ((b)->name_len)
 #define bam_map_qual(b)  ((b)->map_qual)
-#define bam_bin(b)       ((b)->bin)
+#define bam_bin(b)       ((b)->flag & BAM_CIGAR32 ? 0 : (b)->bin)
 #define bam_flag(b)      ((b)->flag)
 #define bam_seq_len(b)   ((b)->len)
 #define bam_strand(b)    ((bam_flag((b)) & BAM_FREVERSE) != 0)
@@ -219,12 +219,12 @@ typedef struct {
 #define bam_set_mate_ref(b,v)   ((b)->mate_ref = (v))
 #define bam_set_mate_pos(b,v)   ((b)->mate_pos = (v))
 #define bam_set_ins_size(b,v)   ((b)->ins_size = (v))
-#define bam_set_cigar_len(b, v) ((b)->cigar_len = (v))
+#define bam_set_cigar_len(b, v) (((v)>>16) ? (((b)->flag |= BAM_CIGAR32), (b)->bin = ((v)>>16), (b)->cigar_len = (v)&0xffff) : ((b)->cigar_len = (v)))
 #define bam_set_name_len(b,v)   ((b)->name_len = (v))
 #define bam_set_map_qual(b,v)   ((b)->map_qual = (v))
-#define bam_set_bin(b,v)        ((b)->bin = (v));
-#define bam_set_flag(b,v)       ((b)->flag = (v));
-#define bam_set_seq_len(b,v)    ((b)->len = (v));
+#define bam_set_bin(b,v)        (((b)->flag & BAM_CIGAR32) || ((b)->bin = (v)))
+#define bam_set_flag(b,v)       ((b)->flag = (v))
+#define bam_set_seq_len(b,v)    ((b)->len = (v))
 
 #define bam_name(b)      ((char *)(&(b)->data))
 #ifdef ALLOW_UAC
@@ -240,23 +240,27 @@ typedef struct {
 #define round4(v) (((v-1)&~3)+4)
 
 /* BAM flags */
-#define BAM_FPAIRED        1
-#define BAM_FPROPER_PAIR   2
-#define BAM_FUNMAP         4
-#define BAM_FMUNMAP        8
-#define BAM_FREVERSE      16
-#define BAM_FMREVERSE     32
-#define BAM_FREAD1        64
-#define BAM_FREAD2       128
-#define BAM_FSECONDARY   256
-#define BAM_FQCFAIL      512
-#define BAM_FDUP        1024
+#define BAM_FPAIRED           1
+#define BAM_FPROPER_PAIR      2
+#define BAM_FUNMAP            4
+#define BAM_FMUNMAP           8
+#define BAM_FREVERSE         16
+#define BAM_FMREVERSE        32
+#define BAM_FREAD1           64
+#define BAM_FREAD2          128
+#define BAM_FSECONDARY      256
+#define BAM_FQCFAIL         512
+#define BAM_FDUP           1024
+#define BAM_FSUPPLEMENTARY 2048
+
+#define BAM_CIGAR32    32768
 
 /* CIGAR operations, taken from samtools bam.h */
 #define BAM_CIGAR_SHIFT 4
 #define BAM_CIGAR_MASK  ((1 << BAM_CIGAR_SHIFT) - 1)
 
 enum cigar_op {
+    BAM_UNKNOWN=-1,
     BAM_CMATCH=0,
     BAM_CINS=1,
     BAM_CDEL=2,
@@ -268,6 +272,24 @@ enum cigar_op {
     BAM_CBASE_MISMATCH=8
 };
 
+/*
+ * Whether this cigar op marches along ref, seq or both
+ *
+ * Op   Ref   Seq
+ * M    1     1
+ * I    0     1
+ * D    1     0
+ * N    1     0
+ * S    0     1
+ * H    0     0
+ * P    0     0
+ * =    1     1
+ * X    1     1
+ */
+#define BAM_CONSUME_REF(op) ((0x18d>>(op))&1)
+#define BAM_CONSUME_SEQ(op) ((0x193>>(op))&1)
+
+
 /* ----------------------------------------------------------------------
  * Function prototypes
  * We only support reading, so basically we have open, read, close along
@@ -725,6 +747,14 @@ int bam_set_option(bam_file_t *fd, enum bam_option opt, ...);
  */
 int bam_set_voption(bam_file_t *fd, enum bam_option opt, va_list args);
 
+
+
+/*
+ * Signed and unsigned fast functions to act as equiv to sprintf(cp, "%d", i)
+ */
+unsigned char *append_int(unsigned char *cp, int32_t i);
+unsigned char *append_uint(unsigned char *cp, uint32_t i);
+
 #ifdef __cplusplus
 }
 #endif
