@@ -343,7 +343,7 @@ int cram_external_decode_int(cram_slice *slice, cram_codec *c,
     /* Find the external block */
     if (slice->block_by_id) {
 	if (!(b = slice->block_by_id[c->external.content_id]))
-	    return -1;
+	    return *out_size?-1:0;
     } else {
 	for (i = 0; i < slice->hdr->num_blocks; i++) {
 	    b = slice->block[i];
@@ -374,7 +374,7 @@ int cram_external_decode_char(cram_slice *slice, cram_codec *c,
     /* Find the external block */
     if (slice->block_by_id) {
 	if (!(b = slice->block_by_id[c->external.content_id]))
-	    return -1;
+	    return *out_size?-1:0;
     } else {
 	for (i = 0; i < slice->hdr->num_blocks; i++) {
 	    b = slice->block[i];
@@ -395,31 +395,18 @@ int cram_external_decode_char(cram_slice *slice, cram_codec *c,
     return 0;
 }
 
-static int cram_external_decode_block2(cram_slice *slice, cram_codec *c,
+static int cram_external_decode_block(cram_slice *slice, cram_codec *c,
 				      cram_block *in, char *out_,
 				      int *out_size) {
-    char *cp;
-    cram_block *b = c->external.b;
-    cram_block *out = (cram_block *)out_;
-
-    cp = cram_extract_block(b, *out_size);
-    if (!cp)
-	return -1;
-
-    BLOCK_APPEND(out, cp, *out_size);
-    return 0;
-}
-
-static int cram_external_decode_block(cram_slice *slice, cram_codec *c,
-				       cram_block *in, char *out_,
-				       int *out_size) {
     int i;
+    char *cp;
+    cram_block *out = (cram_block *)out_;
     cram_block *b = NULL;
 
     /* Find the external block */
     if (slice->block_by_id) {
 	if (!(b = slice->block_by_id[c->external.content_id]))
-	    return -1;
+	    return *out_size?-1:0;
     } else {
 	for (i = 0; i < slice->hdr->num_blocks; i++) {
 	    b = slice->block[i];
@@ -432,10 +419,12 @@ static int cram_external_decode_block(cram_slice *slice, cram_codec *c,
 	    return -1;
     }
 
-    c->external.b = b;
-    c->decode = cram_external_decode_block2;
+    cp = cram_extract_block(b, *out_size);
+    if (!cp)
+	return -1;
 
-    return c->decode(slice, c, in, out_, out_size);
+    BLOCK_APPEND(out, cp, *out_size);
+    return 0;
 }
 
 void cram_external_decode_free(cram_codec *c) {
@@ -475,16 +464,16 @@ cram_codec *cram_external_decode_init(char *data, int size,
 }
 
 int cram_external_encode_int(cram_slice *slice, cram_codec *c,
-			     cram_block *out, char *in, int in_size) {
+			     char *in, int in_size) {
     uint32_t *i32 = (uint32_t *)in;
 
-    itf8_put_blk(out, *i32);
+    itf8_put_blk(c->out, *i32);
     return 0;
 }
 
 int cram_external_encode_char(cram_slice *slice, cram_codec *c,
-			      cram_block *out, char *in, int in_size) {
-    BLOCK_APPEND(out, in, in_size);
+			      char *in, int in_size) {
+    BLOCK_APPEND(c->out, in, in_size);
     return 0;
 }
 
@@ -625,23 +614,25 @@ int cram_beta_encode_store(cram_codec *c, cram_block *b,
 }
 
 int cram_beta_encode_int(cram_slice *slice, cram_codec *c,
-			 cram_block *out, char *in, int in_size) {
+			 char *in, int in_size) {
     int *syms = (int *)in;
     int i, r = 0;
 
     for (i = 0; i < in_size; i++)
-	r |= store_bits_MSB(out, syms[i] + c->e_beta.offset, c->e_beta.nbits);
+	r |= store_bits_MSB(c->out, syms[i] + c->e_beta.offset,
+			    c->e_beta.nbits);
 
     return r;
 }
 
 int cram_beta_encode_char(cram_slice *slice, cram_codec *c,
-			 cram_block *out, char *in, int in_size) {
+			 char *in, int in_size) {
     unsigned char *syms = (unsigned char *)in;
     int i, r = 0;
 
     for (i = 0; i < in_size; i++)
-	r |= store_bits_MSB(out, syms[i] + c->e_beta.offset, c->e_beta.nbits);
+	r |= store_bits_MSB(c->out, syms[i] + c->e_beta.offset,
+			    c->e_beta.nbits);
 
     return r;
 }
@@ -1083,12 +1074,12 @@ cram_codec *cram_huffman_decode_init(char *data, int size,
 }
 
 int cram_huffman_encode_char0(cram_slice *slice, cram_codec *c,
-			      cram_block *out, char *in, int in_size) {
+			      char *in, int in_size) {
     return 0;
 }
 
 int cram_huffman_encode_char(cram_slice *slice, cram_codec *c,
-			     cram_block *out, char *in, int in_size) {
+			     char *in, int in_size) {
     int i, code, len, r = 0;
     unsigned char *syms = (unsigned char *)in;
 
@@ -1112,19 +1103,19 @@ int cram_huffman_encode_char(cram_slice *slice, cram_codec *c,
 	    len  = c->e_huffman.codes[i].len;
 	}
 
-	r |= store_bits_MSB(out, code, len);
+	r |= store_bits_MSB(c->out, code, len);
     } while (--in_size);
 
     return r;
 }
 
 int cram_huffman_encode_int0(cram_slice *slice, cram_codec *c,
-			     cram_block *out, char *in, int in_size) {
+			     char *in, int in_size) {
     return 0;
 }
 
 int cram_huffman_encode_int(cram_slice *slice, cram_codec *c,
-			    cram_block *out, char *in, int in_size) {
+			    char *in, int in_size) {
     int i, code, len, r = 0;
     int *syms = (int *)in;
 
@@ -1149,7 +1140,7 @@ int cram_huffman_encode_int(cram_slice *slice, cram_codec *c,
 	    len  = c->e_huffman.codes[i].len;
 	}
 
-	r |= store_bits_MSB(out, code, len);
+	r |= store_bits_MSB(c->out, code, len);
     } while (--in_size);
 
     return r;
@@ -1460,19 +1451,37 @@ cram_codec *cram_byte_array_len_decode_init(char *data, int size,
 }
 
 int cram_byte_array_len_encode(cram_slice *slice, cram_codec *c,
-			cram_block *out, char *in, int in_size) {
-    return -1; // not imp.
+			       char *in, int in_size) {
+    int32_t i32 = in_size;
+    int r = 0;
+
+    r |= c->e_byte_array_len.len_codec->encode(slice,
+					       c->e_byte_array_len.len_codec,
+					       (char *)&i32, 1);
+    r |= c->e_byte_array_len.val_codec->encode(slice,
+					       c->e_byte_array_len.val_codec,
+					       in, in_size);
+    return r;
 }
 
 void cram_byte_array_len_encode_free(cram_codec *c) {
     if (!c)
 	return;
+
+    if (c->e_byte_array_len.len_codec)
+	c->e_byte_array_len.len_codec->free(c->e_byte_array_len.len_codec);
+
+    if (c->e_byte_array_len.val_codec)
+	c->e_byte_array_len.val_codec->free(c->e_byte_array_len.val_codec);
+
     free(c);
 }
 
 int cram_byte_array_len_encode_store(cram_codec *c, cram_block *b,
 				     char *prefix, int version) {
-    int len = 0;
+    int len = 0, len2, len3;
+    cram_codec *tc;
+    cram_block *b_len, *b_val;
 
     if (prefix) {
 	size_t l = strlen(prefix);
@@ -1480,16 +1489,23 @@ int cram_byte_array_len_encode_store(cram_codec *c, cram_block *b,
 	len += l;
     }
 
+    tc = c->e_byte_array_len.len_codec; 
+    b_len = cram_new_block(0, 0);
+    len2 = tc->store(tc, b_len, NULL, version);
+
+    tc = c->e_byte_array_len.val_codec;
+    b_val = cram_new_block(0, 0);
+    len3 = tc->store(tc, b_val, NULL, version);
+
     len += itf8_put_blk(b, c->codec);
-    len += itf8_put_blk(b, c->e_byte_array_len.len_len +
-			   c->e_byte_array_len.val_len);
-    BLOCK_APPEND(b, c->e_byte_array_len.len_dat, c->e_byte_array_len.len_len);
-    len += c->e_byte_array_len.len_len;
+    len += itf8_put_blk(b, len2+len3);
+    BLOCK_APPEND(b, BLOCK_DATA(b_len), BLOCK_SIZE(b_len));
+    BLOCK_APPEND(b, BLOCK_DATA(b_val), BLOCK_SIZE(b_val));
 
-    BLOCK_APPEND(b, c->e_byte_array_len.val_dat, c->e_byte_array_len.val_len);
-    len += c->e_byte_array_len.val_len;
+    cram_free_block(b_len);
+    cram_free_block(b_val);
 
-    return len;
+    return len + len2 + len3;
 }
 
 cram_codec *cram_byte_array_len_encode_init(cram_stats *st,
@@ -1507,10 +1523,14 @@ cram_codec *cram_byte_array_len_encode_init(cram_stats *st,
     c->encode = cram_byte_array_len_encode;
     c->store = cram_byte_array_len_encode_store;
 
-    c->e_byte_array_len.len_len = e->len_len;
-    c->e_byte_array_len.len_dat = e->len_dat;
-    c->e_byte_array_len.val_len = e->val_len;
-    c->e_byte_array_len.val_dat = e->val_dat;
+    c->e_byte_array_len.len_codec = cram_encoder_init(e->len_encoding,
+						      NULL, E_INT, 
+						      e->len_dat,
+						      version);
+    c->e_byte_array_len.val_codec = cram_encoder_init(e->val_encoding,
+						      NULL, E_BYTE_ARRAY, 
+						      e->val_dat,
+						      version);
 
     return c;
 }
@@ -1519,11 +1539,27 @@ cram_codec *cram_byte_array_len_encode_init(cram_stats *st,
  * ---------------------------------------------------------------------------
  * BYTE_ARRAY_STOP
  */
-static int cram_byte_array_stop_decode_char2(cram_slice *slice, cram_codec *c,
-					     cram_block *in, char *out,
-					     int *out_size) {
-    cram_block *b = c->byte_array_stop.b;
+static int cram_byte_array_stop_decode_char(cram_slice *slice, cram_codec *c,
+					    cram_block *in, char *out,
+					    int *out_size) {
+    int i;
     char *cp, ch;
+    cram_block *b = NULL;
+
+    if (slice->block_by_id) {
+	if (!(b = slice->block_by_id[c->byte_array_stop.content_id]))
+	    return *out_size?-1:0;
+    } else {
+	for (i = 0; i < slice->hdr->num_blocks; i++) {
+	    b = slice->block[i];
+	    if (b && b->content_type == EXTERNAL &&
+		b->content_id == c->byte_array_stop.content_id) {
+		break;
+	    }
+	}
+	if (i == slice->hdr->num_blocks || !b)
+	    return -1;
+    }
 
     if (b->idx >= b->uncomp_size)
 	return -1;
@@ -1542,36 +1578,6 @@ static int cram_byte_array_stop_decode_char2(cram_slice *slice, cram_codec *c,
     return 0;
 }
 
-static int cram_byte_array_stop_decode_char(cram_slice *slice, cram_codec *c,
-					    cram_block *in, char *out,
-					    int *out_size) {
-    int i;
-    cram_block *b = NULL;
-
-    if (slice->block_by_id) {
-	if (!(b = slice->block_by_id[c->byte_array_stop.content_id]))
-	    return -1;
-    } else {
-	for (i = 0; i < slice->hdr->num_blocks; i++) {
-	    b = slice->block[i];
-	    if (b && b->content_type == EXTERNAL &&
-		b->content_id == c->byte_array_stop.content_id) {
-		break;
-	    }
-	}
-	if (i == slice->hdr->num_blocks || !b)
-	    return -1;
-    }
-
-    if (b->idx >= b->uncomp_size)
-	return -1;
-
-    c->byte_array_stop.b = b;
-    c->decode = cram_byte_array_stop_decode_char2;
-
-    return c->decode(slice, c, in, out, out_size);
-}
-
 int cram_byte_array_stop_decode_block(cram_slice *slice, cram_codec *c,
 				      cram_block *in, char *out_,
 				      int *out_size) {
@@ -1582,7 +1588,7 @@ int cram_byte_array_stop_decode_block(cram_slice *slice, cram_codec *c,
 
     if (slice->block_by_id) {
 	if (!(b = slice->block_by_id[c->byte_array_stop.content_id]))
-	    return -1;
+	    return *out_size?-1:0;
     } else {
 	int i;
 	for (i = 0; i < slice->hdr->num_blocks; i++) {
@@ -1661,8 +1667,10 @@ cram_codec *cram_byte_array_stop_decode_init(char *data, int size,
 }
 
 int cram_byte_array_stop_encode(cram_slice *slice, cram_codec *c,
-				cram_block *out, char *in, int in_size) {
-    return -1; // not imp.
+				char *in, int in_size) {
+    BLOCK_APPEND(c->out, in, in_size);
+    BLOCK_APPEND_CHAR(c->out, c->e_byte_array_stop.stop);
+    return 0;
 }
 
 void cram_byte_array_stop_encode_free(cram_codec *c) {
@@ -1796,7 +1804,10 @@ cram_codec *cram_encoder_init(enum cram_encoding codec,
 	return NULL;
 
     if (encode_init[codec]) {
-	return encode_init[codec](st, option, dat, version);
+	cram_codec *r;
+	if ((r = encode_init[codec](st, option, dat, version)))
+	    r->out = NULL;
+	return r;
     } else {
 	fprintf(stderr, "Unimplemented codec of type %s\n", codec2str(codec));
 	abort();
