@@ -472,6 +472,12 @@ cram_block_compression_hdr *cram_decode_compression_header(cram_fd *fd,
 		cram_free_compression_header(hdr);
 		return NULL;
 	    }
+	} else if (key[0] == 'B' && key[1] == 'B') {
+	    if (!(hdr->codecs[DS_BB] = cram_decoder_init(encoding, cp, size, E_BYTE_ARRAY,
+						    fd->version))) {
+		cram_free_compression_header(hdr);
+		return NULL;
+	    }
 	} else if (key[0] == 'R' && key[1] == 'S') {
 	    if (!(hdr->codecs[DS_RS] = cram_decoder_init(encoding, cp, size, E_INT,
 						    fd->version))) {
@@ -509,8 +515,8 @@ cram_block_compression_hdr *cram_decode_compression_header(cram_fd *fd,
 		cram_free_compression_header(hdr);
 		return NULL;
 	    }
-	    if (!(hdr->codecs[DS_Qs] = cram_decoder_init(encoding, cp, size,
-						    E_BYTE_ARRAY,
+	} else if (key[0] == 'Q' && key[1] == 'Q') {
+	    if (!(hdr->codecs[DS_QQ] = cram_decoder_init(encoding, cp, size, E_BYTE_ARRAY,
 						    fd->version))) {
 		cram_free_compression_header(hdr);
 		return NULL;
@@ -600,7 +606,7 @@ int cram_dependent_data_series(cram_fd *fd,
 	DS_BF, DS_AP, DS_FP, DS_RL, DS_DL, DS_NF, DS_BA, DS_QS,
 	DS_FC, DS_FN, DS_BS, DS_IN, DS_RG, DS_MQ, DS_TL, DS_RN,
 	DS_NS, DS_NP, DS_TS, DS_MF, DS_CF, DS_RI, DS_RS, DS_PD,
-	DS_HC, DS_SC
+	DS_HC, DS_SC, DS_BB, DS_QQ,
     };
     uint32_t orig_ds;
 
@@ -608,7 +614,7 @@ int cram_dependent_data_series(cram_fd *fd,
      * Set the data_series bit field based on fd->required_fields
      * contents.
      */
-    if (fd->required_fields) {
+    if (fd->required_fields && fd->required_fields != INT_MAX) {
 	hdr->data_series = 0;
 
 	if (fd->required_fields & SAM_QNAME)
@@ -618,10 +624,10 @@ int cram_dependent_data_series(cram_fd *fd,
 	    hdr->data_series |= CRAM_BF;
 
 	if (fd->required_fields & SAM_RNAME)
-	    hdr->data_series |= CRAM_RI;
+	    hdr->data_series |= CRAM_RI | CRAM_BF;
 
 	if (fd->required_fields & SAM_POS)
-	    hdr->data_series |= CRAM_AP;
+	    hdr->data_series |= CRAM_AP | CRAM_BF;
 
 	if (fd->required_fields & SAM_MAPQ)
 	    hdr->data_series |= CRAM_MQ;
@@ -633,7 +639,7 @@ int cram_dependent_data_series(cram_fd *fd,
 	    hdr->data_series |= CRAM_CF | CRAM_NF | CRAM_RI | CRAM_NS |CRAM_BF;
 
 	if (fd->required_fields & SAM_PNEXT)
-	    hdr->data_series |= CRAM_CF | CRAM_NF | CRAM_AP | CRAM_NP;
+	    hdr->data_series |= CRAM_CF | CRAM_NF | CRAM_AP | CRAM_NP | CRAM_BF;
 
 	if (fd->required_fields & SAM_TLEN)
 	    hdr->data_series |= CRAM_CF | CRAM_NF | CRAM_AP | CRAM_TS |
@@ -647,6 +653,13 @@ int cram_dependent_data_series(cram_fd *fd,
 
 	if (fd->required_fields & SAM_AUX)
 	    hdr->data_series |= CRAM_RG | CRAM_TL | CRAM_aux;
+
+	if (fd->required_fields & SAM_RGAUX)
+	    hdr->data_series |= CRAM_RG | CRAM_BF;
+
+	// Always uncompress CORE block
+	if (cram_uncompress_block(s->block[0]))
+	    return -1;
     } else {
 	hdr->data_series = CRAM_ALL;
 
@@ -671,8 +684,21 @@ int cram_dependent_data_series(cram_fd *fd,
 	 * It's not reciprocal though. We may be needing to decode FN
 	 * but have no need to decode FC, FP and cigar ops.
 	 */
-	if (hdr->data_series & CRAM_SEQ)   hdr->data_series |= CRAM_CIGAR;
-	if (hdr->data_series & CRAM_CIGAR) hdr->data_series |= CRAM_CIGAR;
+	if (hdr->data_series & CRAM_RS)    hdr->data_series |= CRAM_FC|CRAM_FP;
+	if (hdr->data_series & CRAM_PD)    hdr->data_series |= CRAM_FC|CRAM_FP;
+	if (hdr->data_series & CRAM_HC)    hdr->data_series |= CRAM_FC|CRAM_FP;
+	if (hdr->data_series & CRAM_QS)    hdr->data_series |= CRAM_FC|CRAM_FP;
+	if (hdr->data_series & CRAM_IN)    hdr->data_series |= CRAM_FC|CRAM_FP;
+	if (hdr->data_series & CRAM_SC)    hdr->data_series |= CRAM_FC|CRAM_FP;
+	if (hdr->data_series & CRAM_BS)    hdr->data_series |= CRAM_FC|CRAM_FP;
+	if (hdr->data_series & CRAM_DL)    hdr->data_series |= CRAM_FC|CRAM_FP;
+	if (hdr->data_series & CRAM_BA)    hdr->data_series |= CRAM_FC|CRAM_FP;
+	if (hdr->data_series & CRAM_BB)    hdr->data_series |= CRAM_FC|CRAM_FP;
+	if (hdr->data_series & CRAM_QQ)    hdr->data_series |= CRAM_FC|CRAM_FP;
+
+	// cram_decode_seq() needs seq[] array
+	if (hdr->data_series & (CRAM_SEQ|CRAM_CIGAR)) hdr->data_series |= CRAM_RL;
+
 	if (hdr->data_series & CRAM_FP)    hdr->data_series |= CRAM_FC;
 	if (hdr->data_series & CRAM_FC)    hdr->data_series |= CRAM_FN;
 	if (hdr->data_series & CRAM_aux)   hdr->data_series |= CRAM_TL;
@@ -683,7 +709,7 @@ int cram_dependent_data_series(cram_fd *fd,
 	    hdr->data_series |= CRAM_CF;
 	if (!hdr->read_names_included && hdr->data_series & CRAM_RN)
 	    hdr->data_series |= CRAM_CF | CRAM_NF;
-	if (hdr->data_series & (CRAM_BA | CRAM_QS))
+	if (hdr->data_series & (CRAM_BA | CRAM_QS | CRAM_BB | CRAM_QQ))
 	    hdr->data_series |= CRAM_BF | CRAM_CF | CRAM_RL;
 
 	orig_ds = hdr->data_series;
@@ -991,9 +1017,10 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
     uint32_t *cigar = s->cigar;
     uint32_t ncigar = s->ncigar;
     uint32_t cigar_alloc = s->cigar_alloc;
-    uint32_t nm = 0, md_dist = 0;
+    uint32_t nm = 0;
+    int32_t md_dist = 0;
     int orig_aux = 0;
-    int decode_md = fd->decode_md;
+    int decode_md = fd->decode_md && s->ref;
     char buf[20];
     uint32_t ds = c->comp_hdr->data_series;
 
@@ -1016,6 +1043,10 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 
     ref_pos--; // count from 0
     cr->cigar = ncigar;
+
+    if (!(ds & (CRAM_FC | CRAM_FP)))
+	goto skip_cigar;
+
     for (f = 0; f < fn; f++) {
 	int32_t pos = 0;
 	char op;
@@ -1036,7 +1067,7 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 	}
 
 	if (!(ds & CRAM_FP))
-	    goto skip_cigar;
+	    continue;
 
 	if (!c->comp_hdr->codecs[DS_FP]) return -1;
 	r |= c->comp_hdr->codecs[DS_FP]->decode(s,
@@ -1076,12 +1107,16 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 #endif
 	    cig_len += pos - seq_pos;
 	    ref_pos += pos - seq_pos;
-	    md_dist += pos - seq_pos;
+	    if (md_dist >= 0)
+		md_dist += pos - seq_pos;
 	    seq_pos = pos;
 	}
 
 	prev_pos = pos;
 
+	if (!(ds & CRAM_FC))
+	    continue;
+
 	switch(op) {
 	case 'S': { // soft clip: IN
 	    int32_t out_sz2 = 1;
@@ -1091,18 +1126,30 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		cig_len = 0;
 	    }
 	    if (ds & CRAM_IN) {
-		if (IS_CRAM_1_VERS(fd)) {
+		switch (CRAM_MAJOR_VERS(fd->version)) {
+		case 1:
 		    r |= c->comp_hdr->codecs[DS_IN]
 			? c->comp_hdr->codecs[DS_IN]
 			             ->decode(s, c->comp_hdr->codecs[DS_IN],
 					      blk, &seq[pos-1], &out_sz2)
 			: (seq[pos-1] = 'N', out_sz2 = 1, 0);
-		} else {
+		    break;
+
+		case 2:
+		default:
 		    r |= c->comp_hdr->codecs[DS_SC]
 			? c->comp_hdr->codecs[DS_SC]
 			             ->decode(s, c->comp_hdr->codecs[DS_SC],
 					      blk, &seq[pos-1], &out_sz2)
 			: (seq[pos-1] = 'N', out_sz2 = 1, 0);
+		    break;
+
+//		default:
+//		    r |= c->comp_hdr->codecs[DS_BB]
+//			? c->comp_hdr->codecs[DS_BB]
+//			             ->decode(s, c->comp_hdr->codecs[DS_BB],
+//					      blk, &seq[pos-1], &out_sz2)
+//			: (seq[pos-1] = 'N', out_sz2 = 1, 0);
 		}
 		cigar[ncigar++] = (out_sz2<<4) + BAM_CSOFT_CLIP;
 		cig_op = BAM_CSOFT_CLIP;
@@ -1139,13 +1186,20 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 					 (char *)&base, &out_sz);
 		if (ref_pos >= bfd->ref[cr->ref_id].len || !s->ref) {
 		    seq[pos-1] = 'N';
+		    if (decode_md) {
+			if (md_dist >= 0)
+			    BLOCK_APPEND_UINT(s->aux_blk, md_dist);
+			md_dist = -1;
+			nm--;
+		    }
 		} else {
 		    ref_base = fd->L1[(uc)s->ref[ref_pos - s->ref_start +1]];
 		    seq[pos-1] = c->comp_hdr->
 			substitution_matrix[ref_base][base];
 		    if (decode_md) {
-			BLOCK_APPENDF_2(s->aux_blk, buf, "%d%c", md_dist,
-					s->ref[ref_pos-s->ref_start +1]);
+                        BLOCK_APPEND_UINT(s->aux_blk, md_dist);
+                        BLOCK_APPEND_CHAR(s->aux_blk,
+                                          s->ref[ref_pos-s->ref_start +1]);
 			md_dist = 0;
 		    }
 		}
@@ -1170,16 +1224,35 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		                ->decode(s, c->comp_hdr->codecs[DS_DL], blk,
 					 (char *)&i32, &out_sz);
 		if (decode_md) {
-		    BLOCK_APPENDF_1(s->aux_blk, buf, "%d^", md_dist);
-		    BLOCK_APPEND(s->aux_blk,
-				 &s->ref[ref_pos - s->ref_start +1],
-				 i32);
-		    md_dist = 0;
+		    if (md_dist >= 0)
+			BLOCK_APPEND_UINT(s->aux_blk, md_dist);
+		    if (ref_pos + i32 <= bfd->ref[cr->ref_id].len) {
+			BLOCK_APPEND_CHAR(s->aux_blk, '^');
+			BLOCK_APPEND(s->aux_blk,
+				     &s->ref[ref_pos - s->ref_start +1],
+				     i32);
+			nm += i32;
+			md_dist = 0;
+		    } else {
+			uint32_t dlen;
+			if (bfd->ref[cr->ref_id].len >= ref_pos) {
+			    BLOCK_APPEND_CHAR(s->aux_blk, '^');
+			    BLOCK_APPEND(s->aux_blk,
+					 &s->ref[ref_pos - s->ref_start +1],
+					 bfd->ref[cr->ref_id].len - ref_pos);
+			    dlen = i32 - (bfd->ref[cr->ref_id].len - ref_pos);
+			    nm += i32 - dlen;
+			    BLOCK_APPEND_UINT(s->aux_blk, 0);
+			} else {
+			    dlen = i32;
+			}
+
+			md_dist = -1;
+		    }
 		}
 		cig_op = BAM_CDEL;
 		cig_len += i32;
 		ref_pos += i32;
-		nm      += i32;
 		//printf("  %d: DL = %d (ret %d)\n", f, i32, r);
 	    }
 	    break;
@@ -1226,6 +1299,73 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 	    break;
 	}
 
+	case 'b': { // Several bases
+	    int32_t len = 1;
+
+	    if (cig_len && cig_op != BAM_CMATCH) {
+		cigar[ncigar++] = (cig_len<<4) + cig_op;
+		cig_len = 0;
+	    }
+
+	    if (ds & CRAM_BB) {
+		if (!c->comp_hdr->codecs[DS_BB]) return -1;
+		r |= c->comp_hdr->codecs[DS_BB]
+		    ->decode(s, c->comp_hdr->codecs[DS_BB], blk,
+			     (char *)&seq[pos-1], &len);
+
+		if (decode_md) {
+		    int x;
+		    if (md_dist >= 0)
+			BLOCK_APPEND_UINT(s->aux_blk, md_dist);
+
+		    for (x = 0; x < len; x++) {
+			if (x)
+			    BLOCK_APPEND_UINT(s->aux_blk, 0);
+			if (ref_pos + x >= bfd->ref[cr->ref_id].len || !s->ref) {
+			    md_dist = -1;
+			    break;
+			} else {
+			    BLOCK_APPEND_CHAR(s->aux_blk, s->ref[ref_pos+x-s->ref_start +1]);
+			}
+		    }
+
+		    nm += x;
+		}
+	    }
+
+	    cig_op = BAM_CMATCH;
+
+	    cig_len+=len;
+	    seq_pos+=len;
+	    ref_pos+=len;
+	    //prev_pos+=len;
+	    break;
+	}
+
+	case 'q': { // Several quality values
+	    int32_t len = 1;
+
+	    if (cig_len && cig_op != BAM_CMATCH) {
+		cigar[ncigar++] = (cig_len<<4) + cig_op;
+		cig_len = 0;
+	    }
+
+	    if (ds & CRAM_QQ) {
+		if (!c->comp_hdr->codecs[DS_QQ]) return -1;
+		r |= c->comp_hdr->codecs[DS_QQ]
+		    ->decode(s, c->comp_hdr->codecs[DS_QQ], blk,
+			     (char *)&qual[pos-1], &len);
+	    }
+
+	    cig_op = BAM_CMATCH;
+
+	    cig_len+=len;
+	    seq_pos+=len;
+	    ref_pos+=len;
+	    //prev_pos+=len;
+	    break;
+	}
+
 	case 'B': { // Read base; BA, QS
 #ifdef USE_X
 	    if (cig_len && cig_op != BAM_CBASE_MISMATCH) {
@@ -1243,6 +1383,18 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		r |= c->comp_hdr->codecs[DS_BA]
 		                ->decode(s, c->comp_hdr->codecs[DS_BA], blk,
 					 (char *)&seq[pos-1], &out_sz);
+
+		if (decode_md) {
+		    if (md_dist >= 0)
+			BLOCK_APPEND_UINT(s->aux_blk, md_dist);
+		    if (ref_pos >= bfd->ref[cr->ref_id].len || !s->ref) {
+			md_dist = -1;
+		    } else {
+                        BLOCK_APPEND_CHAR(s->aux_blk, s->ref[ref_pos-s->ref_start +1]);
+			nm++;
+			md_dist = 0;
+		    }
+		}
 	    }
 	    if (ds & CRAM_QS) {
 		if (!c->comp_hdr->codecs[DS_QS]) return -1;
@@ -1330,6 +1482,9 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 	}
     }
 
+    if (!(ds & CRAM_FC))
+	goto skip_cigar;
+
     /* An implement match op for any unaccounted for bases */
     if (ds & CRAM_FN && cr->len >= seq_pos) {
 	if (s->ref) {
@@ -1342,7 +1497,8 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 		memcpy(&seq[seq_pos-1], &s->ref[ref_pos - s->ref_start +1],
 		       cr->len - seq_pos + 1);
 		ref_pos += cr->len - seq_pos + 1;
-		md_dist += cr->len - seq_pos + 1;
+		if (md_dist >= 0)
+		    md_dist += cr->len - seq_pos + 1;
 	    }
 	}
 
@@ -1371,7 +1527,8 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
  skip_cigar:
 
     if ((ds & CRAM_FN) && decode_md) {
-	BLOCK_APPENDF_1(s->aux_blk, buf, "%d", md_dist);
+	if (md_dist >= 0)
+	    BLOCK_APPEND_UINT(s->aux_blk, md_dist);
     }
 
     if (cig_len) {
@@ -1403,9 +1560,9 @@ static int cram_decode_seq(cram_fd *fd, cram_container *c, cram_slice *s,
 	int32_t out_sz2 = cr->len;
 
 	if (ds & CRAM_QS) {
-	    if (!c->comp_hdr->codecs[DS_Qs]) return -1;
-	    r |= c->comp_hdr->codecs[DS_Qs]
-		            ->decode(s, c->comp_hdr->codecs[DS_Qs], blk,
+	    if (!c->comp_hdr->codecs[DS_QS]) return -1;
+	    r |= c->comp_hdr->codecs[DS_QS]
+		            ->decode(s, c->comp_hdr->codecs[DS_QS], blk,
 				     qual, &out_sz2);
 	}
     }
@@ -1547,9 +1704,21 @@ static int cram_decode_aux(cram_container *c, cram_slice *s,
 }
 
 /* Resolve mate pair cross-references between recs within this slice */
-static void cram_decode_slice_xref(cram_slice *s) {
+static void cram_decode_slice_xref(cram_slice *s, int required_fields) {
     int rec;
 
+    if (!(required_fields & (SAM_RNEXT | SAM_PNEXT | SAM_TLEN))) {
+	for (rec = 0; rec < s->hdr->num_records; rec++) {
+	    cram_record *cr = &s->crecs[rec];
+
+	    cr->tlen = 0;
+	    cr->mate_pos = 0;
+	    cr->mate_ref_id = -1;
+	}
+
+	return;
+    }
+
     for (rec = 0; rec < s->hdr->num_records; rec++) {
 	cram_record *cr = &s->crecs[rec];
 
@@ -1573,10 +1742,13 @@ static void cram_decode_slice_xref(cram_slice *s) {
 		    int aleft = cr->apos, aright = cr->aend;
 		    int tlen;
 		    int ref = cr->ref_id;
+		    int left_cnt = 0;
 
 		    do {
 			if (aleft > s->crecs[id2].apos)
-			    aleft = s->crecs[id2].apos;
+			    aleft = s->crecs[id2].apos, left_cnt = 1;
+			else if (aleft == s->crecs[id2].apos)
+			    left_cnt++;
 			if (aright < s->crecs[id2].aend)
 			    aright = s->crecs[id2].aend;
 			if (s->crecs[id2].mate_line == -1) {
@@ -1600,9 +1772,8 @@ static void cram_decode_slice_xref(cram_slice *s) {
 			 * bit flags instead, as a tie breaker.
 			 */
 			if (s->crecs[id2].apos == aleft) {
-			    if (s->crecs[id2].aend != aright)
-				s->crecs[id2].tlen = tlen;
-			    else if (s->crecs[id2].flags & BAM_FREAD1)
+			    if (left_cnt == 1 || 
+				(s->crecs[id2].flags & BAM_FREAD1))
 				s->crecs[id2].tlen = tlen;
 			    else
 				s->crecs[id2].tlen = -tlen;
@@ -1613,9 +1784,8 @@ static void cram_decode_slice_xref(cram_slice *s) {
 			id2 = s->crecs[id2].mate_line;
 			while (id2 != id1) {
 			    if (s->crecs[id2].apos == aleft) {
-				if (s->crecs[id2].aend != aright)
-				    s->crecs[id2].tlen = tlen;
-				else if (s->crecs[id2].flags & BAM_FREAD1)
+				if (left_cnt == 1 || 
+				    (s->crecs[id2].flags & BAM_FREAD1))
 				    s->crecs[id2].tlen = tlen;
 				else
 				    s->crecs[id2].tlen = -tlen;
@@ -1750,10 +1920,11 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 	    //s->ref = cram_get_ref(fd, s->hdr->ref_seq_id, 1, 0);
 	    //s->ref_start = 1;
 
-	    s->ref =
-	       cram_get_ref(fd, s->hdr->ref_seq_id,
-	                    s->hdr->ref_seq_start,
-			    s->hdr->ref_seq_start + s->hdr->ref_seq_span -1);
+	    if (fd->required_fields & SAM_SEQ)
+		s->ref =
+		cram_get_ref(fd, s->hdr->ref_seq_id,
+			     s->hdr->ref_seq_start,
+			     s->hdr->ref_seq_start + s->hdr->ref_seq_span -1);
 	    s->ref_start = s->hdr->ref_seq_start;
 	    s->ref_end   = s->hdr->ref_seq_start + s->hdr->ref_seq_span-1;
 
@@ -1764,7 +1935,8 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 	    }
 	    pthread_mutex_lock(&fd->ref_lock);
 	    pthread_mutex_lock(&fd->refs->lock);
-	    if (s->ref_end > fd->refs->ref_id[ref_id]->length) {
+	    if ((fd->required_fields & SAM_SEQ) &&
+		s->ref_end > fd->refs->ref_id[ref_id]->length) {
 		fprintf(stderr, "Slice ends beyond reference end.\n");
 		s->ref_end = fd->refs->ref_id[ref_id]->length;
 	    }
@@ -1773,14 +1945,17 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 	}
     }
 
-    if (s->ref == NULL && s->hdr->ref_seq_id >= 0 && !fd->no_ref) {
+    if ((fd->required_fields & SAM_SEQ) &&
+	s->ref == NULL && s->hdr->ref_seq_id >= 0 && !fd->no_ref) {
 	fprintf(stderr, "Unable to fetch reference #%d %d..%d\n",
 		s->hdr->ref_seq_id, s->hdr->ref_seq_start,
 		s->hdr->ref_seq_start + s->hdr->ref_seq_span-1);
 	return -1;
     }
 
-    if (!IS_CRAM_1_VERS(fd) && s->hdr->ref_seq_id >= 0
+    if (!IS_CRAM_1_VERS(fd)
+	&& (fd->required_fields & SAM_SEQ)
+	&& s->hdr->ref_seq_id >= 0
 	&& !fd->ignore_md5
 	&& memcmp(s->hdr->md5, "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 16)) {
 	MD5_CTX md5;
@@ -1839,6 +2014,7 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 	    return -1;
     }
 
+    int last_ref_id = -9;
     for (rec = 0; rec < s->hdr->num_records; rec++) {
 	cram_record *cr = &s->crecs[rec];
 
@@ -1857,6 +2033,8 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 		return -1;
 	    bf = fd->bam_flag_swap[bf];
 	    cr->flags = bf;
+	} else {
+	    cr->flags = bf = 0x4; // unmapped
 	}
 
 	if (ds & CRAM_CF) {
@@ -1883,12 +2061,19 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 		r |= c->comp_hdr->codecs[DS_RI]
 		                ->decode(s, c->comp_hdr->codecs[DS_RI], blk,
 					 (char *)&cr->ref_id, &out_sz);
-		if (cr->ref_id >= 0) {
+		if ((fd->required_fields & (SAM_SEQ|SAM_TLEN))
+		    && cr->ref_id >= 0
+		    && cr->ref_id != last_ref_id) {
 		    if (!fd->no_ref) {
 			if (!refs[cr->ref_id])
 			    refs[cr->ref_id] = cram_get_ref(fd, cr->ref_id,
 							    1, 0);
 			s->ref = refs[cr->ref_id];
+
+			if (!fd->unsorted && last_ref_id >= 0 && refs[last_ref_id]) {
+			    cram_ref_decr(fd->refs, last_ref_id);
+			    refs[last_ref_id] = NULL;
+			}
 		    }
 		    s->ref_start = 1;
 		    pthread_mutex_lock(&fd->ref_lock);
@@ -1896,7 +2081,11 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 		    s->ref_end = fd->refs->ref_id[cr->ref_id]->length;
 		    pthread_mutex_unlock(&fd->refs->lock);
 		    pthread_mutex_unlock(&fd->ref_lock);
+
+		    last_ref_id = cr->ref_id;
 		}
+	    } else {
+		cr->ref_id = 0;
 	    }
 	} else {
 	    cr->ref_id = ref_id; // Forced constant in CRAM 1.0
@@ -1949,6 +2138,7 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 	    }
 	}
 
+	cr->mate_pos = 0;
 	cr->mate_line = -1;
 	cr->mate_ref_id = -1;
 	if (ds & CRAM_CF && cf & CRAM_FLAG_DETACHED) {
@@ -2106,9 +2296,9 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
 	    if (ds & CRAM_CF && cf & CRAM_FLAG_PRESERVE_QUAL_SCORES) {
 		out_sz2 = cr->len;
 		if (ds & CRAM_QS) {
-		    if (!c->comp_hdr->codecs[DS_Qs]) return -1;
-		    r |= c->comp_hdr->codecs[DS_Qs]
-			            ->decode(s, c->comp_hdr->codecs[DS_Qs],
+		    if (!c->comp_hdr->codecs[DS_QS]) return -1;
+		    r |= c->comp_hdr->codecs[DS_QS]
+			            ->decode(s, c->comp_hdr->codecs[DS_QS],
 					     blk, qual, &out_sz2);
 		}
 	    } else {
@@ -2132,7 +2322,7 @@ int cram_decode_slice(cram_fd *fd, cram_container *c, cram_slice *s,
     pthread_mutex_unlock(&fd->ref_lock);
 
     /* Resolve mate pair cross-references between recs within this slice */
-    cram_decode_slice_xref(s);
+    cram_decode_slice_xref(s, fd->required_fields);
 
     return r;
 }
@@ -2342,15 +2532,19 @@ static cram_slice *cram_next_slice(cram_fd *fd, cram_container **cp) {
 	c->comp_hdr = cram_decode_compression_header(fd, c->comp_hdr_block);
 	if (!c->comp_hdr)
 	    return NULL;
-	if (!c->comp_hdr->AP_delta) {
+	if (!c->comp_hdr->AP_delta &&
+	    sam_hdr_sort_order(fd->header) != ORDER_COORD) {
 	    pthread_mutex_lock(&fd->ref_lock);
 	    fd->unsorted = 1;
 	    pthread_mutex_unlock(&fd->ref_lock);
 	}
     }
 
-    if ((s = c->slice))
+    if ((s = c->slice)) {
+	c->slice = NULL;
 	cram_free_slice(s);
+	s = NULL;
+    }
 
     if (c->curr_slice == c->max_slice) {
 	cram_free_container(c);
@@ -2388,12 +2582,18 @@ static cram_slice *cram_next_slice(cram_fd *fd, cram_container **cp) {
 
 		    if (c->ref_seq_id != fd->range.refid) {
 			fd->eof = 1;
-			return NULL;
+			cram_free_container(c);
+			fd->ctr = NULL;
+			fd->ooc = 1;
+			break;
 		    }
 
 		    if (c->ref_seq_start > fd->range.end) {
 			fd->eof = 1;
-			return NULL;
+			cram_free_container(c);
+			fd->ctr = NULL;
+			fd->ooc = 1;
+			break;
 		    }
 
 		    if (c->ref_seq_start + c->ref_seq_span-1 <
@@ -2417,7 +2617,8 @@ static cram_slice *cram_next_slice(cram_fd *fd, cram_container **cp) {
 		if (!c->comp_hdr)
 		    return NULL;
 
-		if (!c->comp_hdr->AP_delta) {
+		if (!c->comp_hdr->AP_delta &&
+		    sam_hdr_sort_order(fd->header) != ORDER_COORD) {
 		    pthread_mutex_lock(&fd->ref_lock);
 		    fd->unsorted = 1;
 		    pthread_mutex_unlock(&fd->ref_lock);
@@ -2507,6 +2708,8 @@ static cram_slice *cram_next_slice(cram_fd *fd, cram_container **cp) {
 	c = j->c;
 	s = j->s;
 
+	fd->ctr = c;
+
 	t_pool_delete_result(res, 1);
     }
 
