@@ -368,6 +368,11 @@ cram_block *cram_encode_compression_header(cram_fd *fd, cram_container *c,
 	    return NULL;
 	mc++;
     }
+    if (h->codecs[DS_BB]) {
+	if (-1 == h->codecs[DS_BB]->store(h->codecs[DS_BB], map, "BB", fd->version))
+	    return NULL;
+	mc++;
+    }
     if (h->codecs[DS_MQ]) {
 	if (-1 == h->codecs[DS_MQ]->store(h->codecs[DS_MQ], map, "MQ", fd->version))
 	    return NULL;
@@ -383,8 +388,8 @@ cram_block *cram_encode_compression_header(cram_fd *fd, cram_container *c,
 	    return NULL;
 	mc++;
     }
-    if (h->codecs[DS_Qs]) {
-	if (-1 == h->codecs[DS_Qs]->store(h->codecs[DS_Qs], map, "Qs", fd->version))
+    if (h->codecs[DS_QQ]) {
+	if (-1 == h->codecs[DS_QQ]->store(h->codecs[DS_QQ], map, "QQ", fd->version))
 	    return NULL;
 	mc++;
     }
@@ -490,14 +495,20 @@ cram_block *cram_encode_compression_header(cram_fd *fd, cram_container *c,
 				     "\t"   // stop-byte is also SAM separator
 				     DS_aux_BQ_S,
 				     4);
-		    else if (hi->key[0] == 'B' && 
-			     (hi->key[1] == 'D' || hi->key[1] == 'I'))
+		    else if (hi->key[0] == 'B' && hi->key[1] == 'D')
 			BLOCK_APPEND(map,
 				     "\005" // BYTE_ARRAY_STOP
 				     "\002" // len
 				     "\t"   // stop-byte is also SAM separator
 				     DS_aux_BD_S,
 				     4);
+		    else if (hi->key[0] == 'B' && hi->key[1] == 'I')
+			BLOCK_APPEND(map,
+				     "\005" // BYTE_ARRAY_STOP
+				     "\002" // len
+				     "\t"   // stop-byte is also SAM separator
+				     DS_aux_BI_S,
+				     4);
 		    else if ((hi->key[0] == 'Q' && hi->key[1] == '2') ||
 			     (hi->key[0] == 'U' && hi->key[1] == '2') ||
 			     (hi->key[0] == 'Q' && hi->key[1] == 'T') ||
@@ -586,13 +597,14 @@ cram_block *cram_encode_compression_header(cram_fd *fd, cram_container *c,
 		// after slice header construction). So we use
 		// BYTE_ARRAY_LEN with the length codec being external
 		// too.
-		if (hi->key[0] == 'F' && hi->key[1] == 'Z')
+		if ((hi->key[0] == 'F' && hi->key[1] == 'Z') |
+		    (hi->key[0] == 'Z' && hi->key[1] == 'M'))
 		    BLOCK_APPEND(map,
 				 "\004" // BYTE_ARRAY_LEN
 				 "\006" // length
 				 "\001" // EXTERNAL (len)
 				 "\001" // external-len
-				 DS_aux_S // content-id
+				 DS_aux_FZ_S // content-id
 				 "\001" // EXTERNAL (val)
 				 "\001" // external-len
 				 DS_aux_FZ_S,// content-id
@@ -703,31 +715,26 @@ static int cram_encode_slice_read(cram_fd *fd,
     //printf("BF=0x%x\n", cr->flags);
     //	    bf = cram_flag_swap[cr->flags];
     i32 = fd->cram_flag_swap[cr->flags & 0xfff];
-    r |= h->codecs[DS_BF]->encode(s, h->codecs[DS_BF], s->block[DS_BF],
-				  (char *)&i32, 1);
+    r |= h->codecs[DS_BF]->encode(s, h->codecs[DS_BF], (char *)&i32, 1);
 
     i32 = cr->cram_flags;
-    r |= h->codecs[DS_CF]->encode(s, h->codecs[DS_CF], s->block[DS_CF],
-				  (char *)&i32, 1);
+    r |= h->codecs[DS_CF]->encode(s, h->codecs[DS_CF], (char *)&i32, 1);
 
     if (!IS_CRAM_1_VERS(fd) && s->hdr->ref_seq_id == -2)
-	r |= h->codecs[DS_RI]->encode(s, h->codecs[DS_RI], s->block[DS_RI],
-				      (char *)&cr->ref_id, 1);
+	r |= h->codecs[DS_RI]->encode(s, h->codecs[DS_RI], (char *)&cr->ref_id, 1);
 
-    r |= h->codecs[DS_RL]->encode(s, h->codecs[DS_RL], s->block[DS_RL],
-				  (char *)&cr->len, 1);
+    r |= h->codecs[DS_RL]->encode(s, h->codecs[DS_RL], (char *)&cr->len, 1);
 
     if (c->pos_sorted) {
 	i32 = cr->apos - *last_pos;
-	r |= h->codecs[DS_AP]->encode(s, h->codecs[DS_AP], s->block[DS_AP], (char *)&i32, 1);
+	r |= h->codecs[DS_AP]->encode(s, h->codecs[DS_AP], (char *)&i32, 1);
 	*last_pos = cr->apos;
     } else {
 	i32 = cr->apos;
-	r |= h->codecs[DS_AP]->encode(s, h->codecs[DS_AP], s->block[DS_AP], (char *)&i32, 1);
+	r |= h->codecs[DS_AP]->encode(s, h->codecs[DS_AP], (char *)&i32, 1);
     }
 
-    r |= h->codecs[DS_RG]->encode(s, h->codecs[DS_RG], s->block[DS_RG],
-				  (char *)&cr->rg, 1);
+    r |= h->codecs[DS_RG]->encode(s, h->codecs[DS_RG], (char *)&cr->rg, 1);
 
     if (c->comp_hdr->read_names_included) {
 	// RN codec: Already stored in block[3].
@@ -735,23 +742,22 @@ static int cram_encode_slice_read(cram_fd *fd,
 
     if (cr->cram_flags & CRAM_FLAG_DETACHED) {
 	i32 = cr->mate_flags;
-	r |= h->codecs[DS_MF]->encode(s, h->codecs[DS_MF], s->block[DS_MF],
-				      (char *)&i32, 1);
+	r |= h->codecs[DS_MF]->encode(s, h->codecs[DS_MF], (char *)&i32, 1);
 
 	if (!c->comp_hdr->read_names_included) {
 	    // RN codec: Already stored in block[3].
 	}
 
-	r |= h->codecs[DS_NS]->encode(s, h->codecs[DS_NS], s->block[DS_NS],
+	r |= h->codecs[DS_NS]->encode(s, h->codecs[DS_NS], 
 				      (char *)&cr->mate_ref_id, 1);
 
-	r |= h->codecs[DS_NP]->encode(s, h->codecs[DS_NP], s->block[DS_NP],
+	r |= h->codecs[DS_NP]->encode(s, h->codecs[DS_NP], 
 				      (char *)&cr->mate_pos, 1);
 
-	r |= h->codecs[DS_TS]->encode(s, h->codecs[DS_TS], s->block[DS_TS],
+	r |= h->codecs[DS_TS]->encode(s, h->codecs[DS_TS], 
 				      (char *)&cr->tlen, 1);
     } else if (cr->cram_flags & CRAM_FLAG_MATE_DOWNSTREAM) {
-	r |= h->codecs[DS_NF]->encode(s, h->codecs[DS_NF], s->block[DS_NF],
+	r |= h->codecs[DS_NF]->encode(s, h->codecs[DS_NF], 
 				      (char *)&cr->mate_line, 1);
     }
 
@@ -759,17 +765,14 @@ static int cram_encode_slice_read(cram_fd *fd,
     if (IS_CRAM_1_VERS(fd)) {
 	int j;
 	uc = cr->ntags;
-	r |= h->codecs[DS_TC]->encode(s, h->codecs[DS_TC], s->block[DS_TC],
-				      (char *)&uc, 1);
+	r |= h->codecs[DS_TC]->encode(s, h->codecs[DS_TC], (char *)&uc, 1);
 
 	for (j = 0; j < cr->ntags; j++) {
 	    uint32_t i32 = s->TN[cr->TN_idx + j]; // id
-	    r |= h->codecs[DS_TN]->encode(s, h->codecs[DS_TN], s->block[DS_TN],
-					  (char *)&i32, 1);
+	    r |= h->codecs[DS_TN]->encode(s, h->codecs[DS_TN], (char *)&i32, 1);
 	}
     } else {
-	r |= h->codecs[DS_TL]->encode(s, h->codecs[DS_TL], s->block[DS_TL],
-				      (char *)&cr->TL, 1);
+	r |= h->codecs[DS_TL]->encode(s, h->codecs[DS_TL], (char *)&cr->TL, 1);
     }
 
     // qual
@@ -779,17 +782,15 @@ static int cram_encode_slice_read(cram_fd *fd,
     if (!(cr->flags & BAM_FUNMAP)) {
 	int prev_pos = 0, j;
 
-	r |= h->codecs[DS_FN]->encode(s, h->codecs[DS_FN], s->block[DS_FN],
+	r |= h->codecs[DS_FN]->encode(s, h->codecs[DS_FN],
 				      (char *)&cr->nfeature, 1);
 	for (j = 0; j < cr->nfeature; j++) {
 	    cram_feature *f = &s->features[cr->feature + j];
 
 	    uc = f->X.code;
-	    r |= h->codecs[DS_FC]->encode(s, h->codecs[DS_FC], s->block[DS_FC],
-					  (char *)&uc, 1);
+	    r |= h->codecs[DS_FC]->encode(s, h->codecs[DS_FC], (char *)&uc, 1);
 	    i32 = f->X.pos - prev_pos;
-	    r |= h->codecs[DS_FP]->encode(s, h->codecs[DS_FP], s->block[DS_FP],
-					  (char *)&i32, 1);
+	    r |= h->codecs[DS_FP]->encode(s, h->codecs[DS_FP], (char *)&i32, 1);
 	    prev_pos = f->X.pos;
 
 	    switch(f->X.code) {
@@ -799,30 +800,42 @@ static int cram_encode_slice_read(cram_fd *fd,
 		//fprintf(stderr, "    FC=%c FP=%d base=%d\n", f->X.code, i32, f->X.base);
 		
 		uc = f->X.base;
-		r |= h->codecs[DS_BS]->encode(s, h->codecs[DS_BS], s->block[DS_BS],
+		r |= h->codecs[DS_BS]->encode(s, h->codecs[DS_BS], 
 					      (char *)&uc, 1);
 		break;
 	    case 'S':
-		//seq = DSTRING_STR(s->seqs_ds) + f->S.seq_idx;
-		//r |= h->codecs[DS_SC]->encode(s, h->codecs[DS_SC], core,
-		//			     seq, f->S.len);
+		// Already done
+//		r |= h->codecs[DS_SC]->encode(s, h->codecs[DS_SC],
+//					      BLOCK_DATA(s->soft_blk) + f->S.seq_idx,
+//					      f->S.len);
+
+//		if (IS_CRAM_3_VERS(fd)) {
+//		    r |= h->codecs[DS_BB]->encode(s, h->codecs[DS_BB], 
+//						  BLOCK_DATA(s->seqs_blk) + f->S.seq_idx,
+//						  f->S.len);
+//		}
 		break;
 	    case 'I':
 		//seq = DSTRING_STR(s->seqs_ds) + f->S.seq_idx;
-		//r |= h->codecs[DS_IN]->encode(s, h->codecs[DS_IN], core,
+		//r |= h->codecs[DS_IN]->encode(s, h->codecs[DS_IN], 
 		//			     seq, f->S.len);
+//		if (IS_CRAM_3_VERS(fd)) {
+//		    r |= h->codecs[DS_BB]->encode(s, h->codecs[DS_BB], 
+//						  BLOCK_DATA(s->seqs_blk) + f->I.seq_idx,
+//						  f->I.len);
+//		}
 		break;
 	    case 'i':
 		uc = f->i.base;
-		r |= h->codecs[DS_BA]->encode(s, h->codecs[DS_BA], s->block[DS_BA],
+		r |= h->codecs[DS_BA]->encode(s, h->codecs[DS_BA],
 					      (char *)&uc, 1);
 		//seq = DSTRING_STR(s->seqs_ds) + f->S.seq_idx;
-		//r |= h->codecs[DS_IN]->encode(s, h->codecs[DS_IN], core,
+		//r |= h->codecs[DS_IN]->encode(s, h->codecs[DS_IN], 
 		//			     seq, 1);
 		break;
 	    case 'D':
 		i32 = f->D.len;
-		r |= h->codecs[DS_DL]->encode(s, h->codecs[DS_DL], s->block[DS_DL],
+		r |= h->codecs[DS_DL]->encode(s, h->codecs[DS_DL], 
 					      (char *)&i32, 1);
 		break;
 
@@ -831,37 +844,44 @@ static int cram_encode_slice_read(cram_fd *fd,
 		//		    // that aligns against a non ACGTN reference
 
 		uc  = f->B.base;
-		r |= h->codecs[DS_BA]->encode(s, h->codecs[DS_BA], s->block[DS_BA],
+		r |= h->codecs[DS_BA]->encode(s, h->codecs[DS_BA], 
 					      (char *)&uc, 1);
 
 		//                  Already added
 		//		    uc  = f->B.qual;
-		//		    r |= h->codecs[DS_QS]->encode(s, h->codecs[DS_QS], core,
+		//		    r |= h->codecs[DS_QS]->encode(s, h->codecs[DS_QS], 
 		//					     (char *)&uc, 1);
 		break;
 
+	    case 'b':
+		// string of bases
+		r |= h->codecs[DS_BB]->encode(s, h->codecs[DS_BB], 
+					      BLOCK_DATA(s->seqs_blk) + f->b.seq_idx,
+					      f->b.len);
+		break;
+
 	    case 'Q':
 		//                  Already added
 		//		    uc  = f->B.qual;
-		//		    r |= h->codecs[DS_QS]->encode(s, h->codecs[DS_QS], core,
+		//		    r |= h->codecs[DS_QS]->encode(s, h->codecs[DS_QS], 
 		//					     (char *)&uc, 1);
 		break;
 
 	    case 'N':
 		i32 = f->N.len;
-		r |= h->codecs[DS_RS]->encode(s, h->codecs[DS_RS], s->block[DS_RS],
+		r |= h->codecs[DS_RS]->encode(s, h->codecs[DS_RS],
 					      (char *)&i32, 1);
 		break;
 		    
 	    case 'P':
 		i32 = f->P.len;
-		r |= h->codecs[DS_PD]->encode(s, h->codecs[DS_PD], s->block[DS_PD],
+		r |= h->codecs[DS_PD]->encode(s, h->codecs[DS_PD],
 					      (char *)&i32, 1);
 		break;
 		    
 	    case 'H':
 		i32 = f->H.len;
-		r |= h->codecs[DS_HC]->encode(s, h->codecs[DS_HC], s->block[DS_HC],
+		r |= h->codecs[DS_HC]->encode(s, h->codecs[DS_HC],
 					      (char *)&i32, 1);
 		break;
 		    
@@ -873,11 +893,11 @@ static int cram_encode_slice_read(cram_fd *fd,
 	    }
 	}
 
-	r |= h->codecs[DS_MQ]->encode(s, h->codecs[DS_MQ], s->block[DS_MQ],
+	r |= h->codecs[DS_MQ]->encode(s, h->codecs[DS_MQ],
 				      (char *)&cr->mqual, 1);
     } else {
 	char *seq = (char *)BLOCK_DATA(s->seqs_blk) + cr->seq;
-	r |= h->codecs[DS_BA]->encode(s, h->codecs[DS_BA], s->block[DS_BA], seq, cr->len);
+	r |= h->codecs[DS_BA]->encode(s, h->codecs[DS_BA], seq, cr->len);
     }
 
     return r ? -1 : 0;
@@ -896,13 +916,13 @@ static int cram_compress_slice(cram_fd *fd, cram_slice *s) {
     int method = 1<<GZIP | 1<<GZIP_RLE, methodF = method;
 
     /* Compress the CORE Block too, with minimal zlib level */
-    if (level > 5)
+    if (level > 5 && s->block[0]->uncomp_size > 500)
 	cram_compress_block(fd, s->block[0], NULL, GZIP, 1);
  
     if (fd->use_bz2)
 	method |= 1<<BZIP2;
 
-    if (fd->use_arith)
+    if (fd->use_rans)
 	method |= (1<<RANS0) | (1<<RANS1);
 
     if (fd->use_lzma)
@@ -935,6 +955,13 @@ static int cram_compress_slice(cram_fd *fd, cram_slice *s) {
 	if (cram_compress_block(fd, s->block[DS_QS], fd->m[DS_QS],
 				method, 1))
 	    return -1;
+	if (cram_compress_block(fd, s->block[DS_BA], fd->m[DS_BA],
+				method, 1))
+	    return -1;
+	if (s->block[DS_BB])
+	    if (cram_compress_block(fd, s->block[DS_BB], fd->m[DS_BB],
+				    method, 1))
+	    return -1;
 	for (i = DS_aux; i <= DS_aux_oz; i++) {
 	    if (s->block[i])
 		if (cram_compress_block(fd, s->block[i], fd->m[i],
@@ -945,6 +972,13 @@ static int cram_compress_slice(cram_fd *fd, cram_slice *s) {
 	if (cram_compress_block(fd, s->block[DS_QS], fd->m[DS_QS],
 				method, level))
 	    return -1;
+	if (cram_compress_block(fd, s->block[DS_BA], fd->m[DS_BA],
+				method, level))
+	    return -1;
+	if (s->block[DS_BB])
+	    if (cram_compress_block(fd, s->block[DS_BB], fd->m[DS_BB],
+				    method, level))
+	    return -1;
 	for (i = DS_aux; i <= DS_aux_oz; i++) {
 	    if (s->block[i])
 		if (cram_compress_block(fd, s->block[i], fd->m[i],
@@ -957,7 +991,7 @@ static int cram_compress_slice(cram_fd *fd, cram_slice *s) {
     // It benefits well from a little bit extra compression level.
     if (cram_compress_block(fd, s->block[DS_RN], fd->m[DS_RN],
 			    method & ~(1<<RANS0 | 1<<GZIP_RLE),
-			    MIN(9,level+1)))
+			    MIN(9,level)))
 	return -1;
 
     // NS shows strong local correlation as rearrangements are localised
@@ -1052,35 +1086,53 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
      * All the data-series blocks if appropriate. 
      */
     for (id = DS_BF; id < DS_TN; id++) {
-	if (h->codecs[id] && h->codecs[id]->codec == E_EXTERNAL) {
-	    if (!(s->block[id] = cram_new_block(EXTERNAL, id)))
-		return -1;
-	    h->codecs[id]->external.content_id = id;
-	} else {
-	    s->block[id] = s->block[0];
-	}
-    }
+	if (h->codecs[id] && (h->codecs[id]->codec == E_EXTERNAL ||
+			      h->codecs[id]->codec == E_BYTE_ARRAY_STOP ||
+			      h->codecs[id]->codec == E_BYTE_ARRAY_LEN)) {
+	    switch (h->codecs[id]->codec) {
+	    case E_EXTERNAL:
+		if (!(s->block[id] = cram_new_block(EXTERNAL, id)))
+		    return -1;
+		h->codecs[id]->external.content_id = id;
+		break;
 
-#if 0
-    // RANDOMISER
-    {
-	cram_block *bb[5] = {0};
-	int id, j;
+	    case E_BYTE_ARRAY_STOP:
+		if (!(s->block[id] = cram_new_block(EXTERNAL, id)))
+		    return -1;
+		h->codecs[id]->byte_array_stop.content_id = id;
+		break;
 
-	for (id = DS_BF; id < DS_TN; id++) {
-	    if (!h->codecs[id] || h->codecs[id]->codec != E_EXTERNAL)
-		continue;
-	    if (s->block[id] != s->block[0]) {
-		free(s->block[id]);
-		j = random()%5;
-		if (!bb[j])
-		    bb[j] = cram_new_block(EXTERNAL, j+55);
-		s->block[id] = bb[j];
-		h->codecs[id]->external.content_id = j+55;
+	    case E_BYTE_ARRAY_LEN: {
+		cram_codec *cc;
+
+		cc = h->codecs[id]->e_byte_array_len.len_codec;
+		if (cc->codec == E_EXTERNAL) {
+		    int eid = cc->external.content_id;
+		    if (!(s->block[eid] = cram_new_block(EXTERNAL, eid)))
+			return -1;
+		    cc->external.content_id = eid;
+		    cc->out = s->block[eid];
+		}
+
+		cc = h->codecs[id]->e_byte_array_len.val_codec;
+		if (cc->codec == E_EXTERNAL) {
+		    int eid = cc->external.content_id;
+		    if (!s->block[eid])
+			if (!(s->block[eid] = cram_new_block(EXTERNAL, eid)))
+			    return -1;
+		    cc->external.content_id = eid;
+		    cc->out = s->block[eid];
+		}
+		break;
 	    }
+	    }
+	} else {
+	    if (!(id == DS_BB && !h->codecs[DS_BB]))
+		s->block[id] = s->block[0];
 	}
+	if (h->codecs[id])
+	    h->codecs[id]->out = s->block[id];
     }
-#endif
 
     /* Encode reads */
     last_pos = s->hdr->ref_seq_start;
@@ -1102,6 +1154,7 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
     s->block[DS_aux_OQ]= s->aux_OQ_blk;  s->aux_OQ_blk  = NULL;
     s->block[DS_aux_BQ]= s->aux_BQ_blk;  s->aux_BQ_blk  = NULL;
     s->block[DS_aux_BD]= s->aux_BD_blk;  s->aux_BD_blk  = NULL;
+    s->block[DS_aux_BI]= s->aux_BI_blk;  s->aux_BI_blk  = NULL;
     s->block[DS_aux_FZ]= s->aux_FZ_blk;  s->aux_FZ_blk  = NULL;
     s->block[DS_aux_oq]= s->aux_oq_blk;  s->aux_oq_blk  = NULL;
     s->block[DS_aux_os]= s->aux_os_blk;  s->aux_os_blk  = NULL;
@@ -1120,34 +1173,17 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
     if (cram_compress_slice(fd, s) == -1)
 	return -1;
 
-#if 0
-    // RANDOMISER
-    {
-	int i, j;
-	int b_used[100] = {0};
-	for (i = j = 1; i < DS_END; i++) {
-	    if (!s->block[i] || s->block[i] == s->block[0])
-		continue;
-	    s->block[j] = s->block[i];
-	    s->hdr->block_content_ids[j-1] = s->block[i]->content_id;
-	    if (!b_used[s->block[i]->content_id]) {
-		b_used[s->block[i]->content_id] = 1;
-		j++;
-	    }
-	}
-	s->hdr->num_content_ids = j-1;
-	s->hdr->num_blocks = j;
-
-	if (!(s->hdr_block = cram_encode_slice_header(fd, s)))
-	    return -1;
-    }
-#else
     // Collapse empty blocks and create hdr_block
     {
 	int i, j;
 	for (i = j = 1; i < DS_END; i++) {
 	    if (!s->block[i] || s->block[i] == s->block[0])
 		continue;
+	    if (s->block[i]->uncomp_size == 0) {
+		cram_free_block(s->block[i]);
+		s->block[i] = NULL;
+		continue;
+	    }
 	    s->block[j] = s->block[i];
 	    s->hdr->block_content_ids[j-1] = s->block[i]->content_id;
 	    j++;
@@ -1158,7 +1194,6 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
 	if (!(s->hdr_block = cram_encode_slice_header(fd, s)))
 	    return -1;
     }
-#endif
 
     return r ? -1 : 0;
 }
@@ -1181,19 +1216,19 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
     nref = fd->refs->nref;
     pthread_mutex_unlock(&fd->ref_lock);
 
-    if (c->refs_used) {
+    if (!fd->no_ref && c->refs_used) {
 	for (i = 0; i < nref; i++) {
-	    if (c->refs_used[i]) {	
+	    if (c->refs_used[i])
 		cram_get_ref(fd, i, 1, 0);
-	    }
 	}
     }
-
+    /* To create M5 strings */
     /* Fetch reference sequence */
     if (!fd->no_ref) {
 	bam_seq_t *b = c->bams[0];
 	char *ref;
 
+
 	ref = cram_get_ref(fd, bam_ref(b), 1, 0);
 	if (!ref && bam_ref(b) >= 0) {
 	    fprintf(stderr, "Failed to load reference #%d\n", bam_ref(b));
@@ -1205,10 +1240,12 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
 	    c->ref_start = 1;
 	    c->ref_end   = fd->refs->ref_id[c->ref_seq_id]->length;
 	} else {
-	    c->ref_seq_id = c->ref_id; // FIXME remove one var!
+	    c->ref_seq_id = c->ref_id; // FIXME remove one var?
 	}
     } else {
-	    c->ref_seq_id = c->ref_id; // FIXME remove one var!
+	c->ref_id = bam_ref(c->bams[0]);
+	cram_ref_incr(fd->refs, c->ref_id);
+	c->ref_seq_id = c->ref_id;
     }
 
     /* Turn bams into cram_records and gather basic stats */
@@ -1268,6 +1305,11 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
 	s->hdr->num_records = r2;
     }
 
+    if (c->multi_seq && !fd->no_ref) {
+	if (c->ref_seq_id >= 0)
+	    cram_ref_decr(fd->refs, c->ref_seq_id);
+    }
+
     /* Link our bams[] array onto the spare bam list for reuse */
     spares = malloc(sizeof(*spares));
     pthread_mutex_lock(&fd->bam_list_lock);
@@ -1402,6 +1444,23 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
 				    c->stats[DS_BA], E_BYTE, NULL,
 				    fd->version);
 
+    if (CRAM_MAJOR_VERS(fd->version) >= 3) {
+	cram_byte_array_len_encoder e;
+
+	e.len_encoding = E_EXTERNAL;
+	e.len_dat = (void *)DS_BB_len;
+	//e.len_dat = (void *)DS_BB;
+
+	e.val_encoding = E_EXTERNAL;
+	e.val_dat = (void *)DS_BB;
+
+	h->codecs[DS_BB] = cram_encoder_init(E_BYTE_ARRAY_LEN, NULL,
+					     E_BYTE_ARRAY, (void *)&e,
+					     fd->version);
+    } else {
+	h->codecs[DS_BB] = NULL;
+    }
+
     if (fd->verbose > 1) fprintf(stderr, "BS_stats: ");
     h->codecs[DS_BS] = cram_encoder_init(cram_stats_encoding(fd, c->stats[DS_BS]),
 				    c->stats[DS_BS], E_BYTE, NULL,
@@ -1426,8 +1485,6 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
 					c->stats[DS_TN], E_INT, NULL,
 					fd->version);
     } else {
-	int i2[2] = {0, DS_SC};
-
 	h->codecs[DS_TC] = NULL;
 	h->codecs[DS_TN] = NULL;
 
@@ -1457,9 +1514,28 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
 					c->stats[DS_HC], E_INT, NULL,
 					fd->version);
 	//fprintf(stderr, "=== SC ===\n");
-	h->codecs[DS_SC] = cram_encoder_init(E_BYTE_ARRAY_STOP, NULL,
-					E_BYTE_ARRAY, (void *)i2,
-					fd->version);
+	if (1) {
+	    int i2[2] = {0, DS_SC};
+
+	    h->codecs[DS_SC] = cram_encoder_init(E_BYTE_ARRAY_STOP, NULL,
+						 E_BYTE_ARRAY, (void *)i2,
+						 fd->version);
+	} else {
+	    // Appears to be no practical benefit to using this method,
+	    // but it may work better if we start mixing SC, IN and BB
+	    // elements into the same external block.
+	    cram_byte_array_len_encoder e;
+
+	    e.len_encoding = E_EXTERNAL;
+	    e.len_dat = (void *)DS_SC_len;
+
+	    e.val_encoding = E_EXTERNAL;
+	    e.val_dat = (void *)DS_SC;
+
+	    h->codecs[DS_SC] = cram_encoder_init(E_BYTE_ARRAY_LEN, NULL,
+						 E_BYTE_ARRAY, (void *)&e,
+						 fd->version);
+	}
     }
     
     //fprintf(stderr, "=== IN ===\n");
@@ -1571,7 +1647,7 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
     }
 
     /* Cache references up-front if we have unsorted access patterns */
-    if (c->refs_used) {
+    if (!fd->no_ref && c->refs_used) {
 	for (i = 0; i < fd->refs->nref; i++) {
 	    if (c->refs_used[i])
 		cram_ref_decr(fd->refs, i);
@@ -1640,6 +1716,20 @@ static int cram_add_substitution(cram_fd *fd, cram_container *c,
     return cram_add_feature(c, s, r, &f);
 }
 
+static int cram_add_bases(cram_fd *fd, cram_container *c,
+			  cram_slice *s, cram_record *r,
+			  int pos, int len, char *base) {
+    cram_feature f;
+    int i;
+
+    f.b.pos = pos+1;
+    f.b.code = 'b';
+    f.b.seq_idx = base - (char *)BLOCK_DATA(s->seqs_blk);
+    f.b.len = len;
+
+    return cram_add_feature(c, s, r, &f);
+}
+
 static int cram_add_base(cram_fd *fd, cram_container *c,
 			 cram_slice *s, cram_record *r,
 			 int pos, char base, char qual) {
@@ -1688,11 +1778,15 @@ static int cram_add_softclip(cram_container *c, cram_slice *s, cram_record *r,
     f.S.pos = pos+1;
     f.S.code = 'S';
     f.S.len = len;
-    if (CRAM_MAJOR_VERS(version) == 1) {
+    switch (CRAM_MAJOR_VERS(version)) {
+    case 1:
 	f.S.seq_idx = BLOCK_SIZE(s->base_blk);
 	BLOCK_APPEND(s->base_blk, base, len);
 	BLOCK_APPEND_CHAR(s->base_blk, '\0');
-    } else {
+	break;
+
+    case 2:
+    default:
 	f.S.seq_idx = BLOCK_SIZE(s->soft_blk);
 	if (base) {
 	    BLOCK_APPEND(s->soft_blk, base, len);
@@ -1702,6 +1796,11 @@ static int cram_add_softclip(cram_container *c, cram_slice *s, cram_record *r,
 		BLOCK_APPEND_CHAR(s->soft_blk, 'N');
 	}
 	BLOCK_APPEND_CHAR(s->soft_blk, '\0');
+	break;
+
+//    default:
+//	// v3.0 onwards uses BB data-series
+//	f.S.seq_idx = BLOCK_SIZE(s->soft_blk);
     }
     return cram_add_feature(c, s, r, &f);
 }
@@ -1737,7 +1836,7 @@ static int cram_add_pad(cram_container *c, cram_slice *s, cram_record *r,
 }
 
 static int cram_add_insertion(cram_container *c, cram_slice *s, cram_record *r,
-			      int pos, int len, char *base) {
+			      int pos, int len, char *base, int version) {
     cram_feature f;
     f.I.pos = pos+1;
     if (len == 1) {
@@ -1972,8 +2071,8 @@ static char *cram_encode_aux(cram_fd *fd, bam_seq_t *b, cram_container *c,
 	    continue;
 	}
 
-	// BD:Z and BI:Z
-	if (aux[0] == 'B' && (aux[1]=='D' || aux[1]=='I') && aux[2] == 'Z') {
+	// BD:Z
+	if (aux[0] == 'B' && aux[1]=='D' && aux[2] == 'Z') {
 	    char *tmp;
 	    if (!s->aux_BD_blk)
 		if (!(s->aux_BD_blk = cram_new_block(EXTERNAL, DS_aux_BD)))
@@ -1987,6 +2086,21 @@ static char *cram_encode_aux(cram_fd *fd, bam_seq_t *b, cram_container *c,
 	    continue;
 	}
 
+	// BI:Z
+	if (aux[0] == 'B' && aux[1]=='I' && aux[2] == 'Z') {
+	    char *tmp;
+	    if (!s->aux_BI_blk)
+		if (!(s->aux_BI_blk = cram_new_block(EXTERNAL, DS_aux_BI)))
+		    return NULL;
+	    BLOCK_GROW(s->aux_BI_blk, aux_size*1.34+1);
+	    tmp = (char *)BLOCK_END(s->aux_BI_blk);
+	    aux += 3;
+	    while ((*tmp++=*aux++));
+	    *tmp++ = '\t';
+	    BLOCK_SIZE(s->aux_BI_blk) = (uc *)tmp - BLOCK_DATA(s->aux_BI_blk);
+	    continue;
+	}
+
 	// OQ:Z:
 	if (aux[0] == 'O' && aux[1] == 'Q' && aux[2] == 'Z') {
 	    char *tmp;
@@ -2002,17 +2116,48 @@ static char *cram_encode_aux(cram_fd *fd, bam_seq_t *b, cram_container *c,
 	    continue;
 	}
 
-	// FZ:B
-	if (aux[0] == 'F' && aux[1] == 'Z') {
+	// FZ:B or ZM:B
+	if ((aux[0] == 'F' && aux[1] == 'Z' && aux[2] == 'B') ||
+	    (aux[0] == 'Z' && aux[1] == 'M' && aux[2] == 'B')) {
+	    int type = aux[3], blen;
+	    uint32_t count = (uint32_t)((((unsigned char *)aux)[4]<< 0) +
+					(((unsigned char *)aux)[5]<< 8) +
+					(((unsigned char *)aux)[6]<<16) +
+					(((unsigned char *)aux)[7]<<24));
 	    char *tmp;
 	    if (!s->aux_FZ_blk)
 		if (!(s->aux_FZ_blk = cram_new_block(EXTERNAL, DS_aux_FZ)))
 		    return NULL;
 	    BLOCK_GROW(s->aux_FZ_blk, aux_size*1.34+1);
 	    tmp = (char *)BLOCK_END(s->aux_FZ_blk);
-	    aux += 3;
-	    while ((*tmp++=*aux++));
-	    *tmp++ = '\t';
+
+	    // skip TN field
+	    aux+=3;
+
+	    // We use BYTE_ARRAY_LEN with external length, so store that first
+	    switch (type) {
+	    case 'c': case 'C':
+		blen = count;
+		break;
+	    case 's': case 'S':
+		blen = 2*count;
+		break;
+	    case 'i': case 'I': case 'f':
+		blen = 4*count;
+		break;
+	    default:
+		fprintf(stderr, "Unknown sub-type '%c' for aux type 'B'\n",
+			type);
+		return NULL;
+		    
+	    }
+
+	    blen += 5; // sub-type & length
+	    tmp += itf8_put(tmp, blen);
+
+	    // The tag data itself
+	    memcpy(tmp, aux, blen); tmp += blen; aux += blen;
+
 	    BLOCK_SIZE(s->aux_FZ_blk) = (uc *)tmp - BLOCK_DATA(s->aux_FZ_blk);
 	    continue;
 	}
@@ -2075,6 +2220,7 @@ static char *cram_encode_aux(cram_fd *fd, bam_seq_t *b, cram_container *c,
 	    aux+=3; //*tmp++=*aux++; *tmp++=*aux++; *tmp++=*aux++;
 	    *tmp++=*aux++; *tmp++=*aux++; *tmp++=*aux++; *tmp++=*aux++;
 	    *tmp++=*aux++; *tmp++=*aux++; *tmp++=*aux++; *tmp++=*aux++;
+	    break;
 
 	case 'Z': case 'H': {
 		char *tmp;
@@ -2118,10 +2264,8 @@ static char *cram_encode_aux(cram_fd *fd, bam_seq_t *b, cram_container *c,
 		    
 	    }
 
-	    tmp += itf8_put(tmp, blen+5);
-
-	    *tmp++=*aux++; // sub-type & length
-	    *tmp++=*aux++; *tmp++=*aux++; *tmp++=*aux++; *tmp++=*aux++;
+	    blen += 5; // sub-type & length
+	    tmp += itf8_put(tmp, blen);
 
 	    // The tag data itself
 	    memcpy(tmp, aux, blen); tmp += blen; aux += blen;
@@ -2277,7 +2421,7 @@ static cram_container *cram_next_container(cram_fd *fd, bam_seq_t *b) {
 static int process_one_read(cram_fd *fd, cram_container *c,
 			    cram_slice *s, cram_record *cr,
 			    bam_seq_t *b, int rnum) {
-    int i, fake_qual = 0;
+    int i, fake_qual = -1;
     char *cp, *rg;
     char *ref, *seq, *qual;
 
@@ -2321,7 +2465,9 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 	cr->flags |= BAM_FUNMAP;
     cram_stats_add(c->stats[DS_BF], fd->cram_flag_swap[cr->flags & 0xfff]);
 
-    if (!fd->no_ref)
+    // Non reference based encoding means storing the bases verbatim as features, which in
+    // turn means every base also has a quality already stored.
+    if (!fd->no_ref || CRAM_MAJOR_VERS(fd->version) >= 3)
 	cr->cram_flags = CRAM_FLAG_PRESERVE_QUAL_SCORES;
     else
 	cr->cram_flags = 0;
@@ -2366,7 +2512,7 @@ static int process_one_read(cram_fd *fd, cram_container *c,
     /*
      * This seqs_ds is largely pointless and it could reuse the same memory
      * over and over.
-     * s->base_ds is what we need for encoding.
+     * s->base_blk is what we need for encoding.
      */
     cr->seq         = BLOCK_SIZE(s->seqs_blk);
     cr->qual        = BLOCK_SIZE(s->qual_blk);
@@ -2476,9 +2622,29 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 			if (rp[l] != sp[l]) {
 			    if (!sp[l])
 				break;
-			    if (cram_add_substitution(fd, c, s, cr, spos+l,
-						      sp[l], qp[l], rp[l]))
-				return -1;
+			    if (0 && IS_CRAM_3_VERS(fd)) {
+				// Disabled for the time being as it doesn't
+				// seem to gain us much.
+				int ol=l;
+				while (l<end && rp[l] != sp[l])
+				    l++;
+				if (l-ol > 1) {
+				    if (cram_add_bases(fd, c, s, cr, spos+ol,
+						       l-ol, &seq[spos+ol]))
+					return -1;
+				    l--;
+				} else {
+				    l = ol;
+				    if (cram_add_substitution(fd, c, s, cr,
+							      spos+l, sp[l],
+							      qp[l], rp[l]))
+					return -1;
+				}
+			    } else {
+				if (cram_add_substitution(fd, c, s, cr, spos+l,
+							  sp[l], qp[l], rp[l]))
+				    return -1;
+			    }
 			}
 		    }
 		    spos += l;
@@ -2486,11 +2652,26 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 		}
 
 		if (l < cig_len && cr->len) {
-		    /* off end of sequence or non-ref based output */
-		    for (; l < cig_len && seq[spos]; l++, spos++) {
-			if (cram_add_base(fd, c, s, cr, spos,
-					  seq[spos], qual[spos]))
-			    return -1;
+		    if (fd->no_ref) {
+			if (IS_CRAM_3_VERS(fd)) {
+			    if (cram_add_bases(fd, c, s, cr, spos,
+					       cig_len-l, &seq[spos]))
+				return -1;
+			    spos += cig_len-l;
+			} else {
+			    for (; l < cig_len && seq[spos]; l++, spos++) {
+				if (cram_add_base(fd, c, s, cr, spos,
+						  seq[spos], qual[spos]))
+				    return -1;
+			    }
+			}
+		    } else {
+			/* off end of sequence or non-ref based output */
+			for (; l < cig_len && seq[spos]; l++, spos++) {
+			    if (cram_add_base(fd, c, s, cr, spos,
+					      seq[spos], qual[spos]))
+				return -1;
+			}
 		    }
 		    apos += cig_len;
 		} else if (!cr->len) {
@@ -2514,7 +2695,8 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 
 	    case BAM_CINS:
 		if (cram_add_insertion(c, s, cr, spos, cig_len,
-				       cr->len ? &seq[spos] : NULL))
+				       cr->len ? &seq[spos] : NULL,
+				       fd->version))
 		    return -1;
 		if (fd->no_ref && cr->len) {
 		    for (l = 0; l < cig_len; l++, spos++) {
@@ -2530,7 +2712,9 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 				      cr->len ? &seq[spos] : NULL,
 				      fd->version))
 		    return -1;
-		if (fd->no_ref) {
+
+		if (fd->no_ref &&
+		    !(cr->cram_flags & CRAM_FLAG_PRESERVE_QUAL_SCORES)) {
 		    if (cr->len) {
 			for (l = 0; l < cig_len; l++, spos++) {
 			    cram_add_quality(fd, c, s, cr, spos, qual[spos]);
@@ -2600,7 +2784,8 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 	BLOCK_SIZE(s->qual_blk) += cr->len;
     } else {
 	if (cr->len == 0) {
-	    cram_stats_add(c->stats[DS_RL], cr->len = cr->aend - cr->apos + 1);
+	    cr->len = fake_qual >= 0 ? fake_qual : cr->aend - cr->apos + 1;
+	    cram_stats_add(c->stats[DS_RL], cr->len);
 	}
     }
 
@@ -2614,7 +2799,7 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 	//fprintf(stderr, "Checking %"PRId64"/%.*s\t", hd.i,
 	//	cr->name_len, DSTRING_STR(s->name_ds)+cr->name);
 	if (cr->flags & BAM_FPAIRED) {
-	    hi = HashTableAdd(s->pair,
+	    hi = HashTableAdd(s->pair[(cr->flags & BAM_FSECONDARY) ? 1 : 0],
 			      (char *)BLOCK_DATA(s->name_blk)+cr->name,
 			      cr->name_len, hd, &new);
 	    if (!hi)
@@ -2623,40 +2808,94 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 	    new = 1;
 	}
 
+	//printf("%d..%d %.*s new=%d, ins_sz=%d\n",
+	//       cr->apos, cr->aend,
+	//       cr->name_len, (char *)BLOCK_DATA(s->name_blk)+cr->name, new, bam_ins_size(b));
+
 	if (!new) {
 	    cram_record *p = &s->crecs[hi->data.i];
-	    
-	    //fprintf(stderr, "paired %"PRId64"\n", hi->data.i);
+	    int aleft, aright, sign;
 
-	    // copy from p to cr
+	    aleft = MIN(cr->apos, p->apos);
+	    aright = MAX(cr->aend, p->aend);
+	    if (cr->apos < p->apos) {
+		sign = 1;
+	    } else if (cr->apos > p->apos) {
+		sign = -1;
+	    } else if (cr->flags & BAM_FREAD1) {
+		sign = 1;
+	    } else {
+		sign = -1;
+	    }
+
+	    // This vs p: tlen, matepos, flags
+	    if (bam_ins_size(b) != sign*(aright-aleft+1))
+		goto detached;
+
+	    if (MAX(bam_mate_pos(b)+1, 0) != p->apos)
+		goto detached;
+
+	    if (((bam_flag(b) & BAM_FMUNMAP) != 0) !=
+		((p->flags & BAM_FUNMAP) != 0))
+		goto detached;
+
+	    if (((bam_flag(b) & BAM_FMREVERSE) != 0) !=
+		((p->flags & BAM_FREVERSE) != 0))
+		goto detached;
+
+
+	    // p vs this: tlen, matepos, flags
+	    if (p->tlen != -sign*(aright-aleft+1))
+		goto detached;
+
+	    if (p->mate_pos != cr->apos)
+		goto detached;
+
+	    if (((p->flags & BAM_FMUNMAP) != 0) !=
+		((p->mate_flags & CRAM_M_UNMAP) != 0))
+		goto detached;
+
+	    if (((p->flags & BAM_FMREVERSE) != 0) !=
+		((p->mate_flags & CRAM_M_REVERSE) != 0))
+		goto detached;
+
+	    // Supplementary reads are just too ill defined
+	    if ((cr->flags & BAM_FSUPPLEMENTARY) ||
+		(p->flags & BAM_FSUPPLEMENTARY))
+		goto detached;
+
+	    /*
+	     * The fields below are unused when encoding this read as it is
+	     * no longer detached.  In theory they may get referred to when
+	     * processing a 3rd or 4th read in this template?, so we set them
+	     * here just to be sure.
+	     *
+	     * They do not need cram_stats_add() calls those as they are
+	     * not emitted.
+	     */
 	    cr->mate_pos = p->apos;
-	    cram_stats_add(c->stats[DS_NP], cr->mate_pos);
-
-	    cr->tlen = cr->aend - p->apos;
-	    cram_stats_add(c->stats[DS_TS], cr->tlen);
-
+	    cr->tlen = sign*(aright-aleft+1);
 	    cr->mate_flags =
-		((p->flags & BAM_FMUNMAP)   == BAM_FMUNMAP)   * CRAM_M_UNMAP +
-		((p->flags & BAM_FMREVERSE) == BAM_FMREVERSE) * CRAM_M_REVERSE;
-	    cram_stats_add(c->stats[DS_MF], cr->mate_flags);
+	    	((p->flags & BAM_FMUNMAP)   == BAM_FMUNMAP)   * CRAM_M_UNMAP +
+	    	((p->flags & BAM_FMREVERSE) == BAM_FMREVERSE) * CRAM_M_REVERSE;
 
-	    // copy from cr to p
+	    // Decrement statistics aggregated earlier
 	    cram_stats_del(c->stats[DS_NP], p->mate_pos);
-	    p->mate_pos = cr->apos;
-	    cram_stats_add(c->stats[DS_NP], p->mate_pos);
-
 	    cram_stats_del(c->stats[DS_MF], p->mate_flags);
-	    p->mate_flags =
-		((cr->flags & BAM_FMUNMAP)   == BAM_FMUNMAP)  * CRAM_M_UNMAP +
-		((cr->flags & BAM_FMREVERSE) == BAM_FMREVERSE)* CRAM_M_REVERSE;
-	    cram_stats_add(c->stats[DS_MF], p->mate_flags);
-
 	    cram_stats_del(c->stats[DS_TS], p->tlen);
-	    p->tlen = p->apos - cr->aend;
-	    cram_stats_add(c->stats[DS_TS], p->tlen);
+	    cram_stats_del(c->stats[DS_NS], p->mate_ref_id);
+
+	    /* Similarly we could correct the p-> values too, but these will no
+	     * longer have any code that refers back to them as the new 'p'
+	     * for this template is our current 'cr'.
+	     */
+	    //p->mate_pos = cr->apos;
+	    //p->mate_flags =
+	    //	((cr->flags & BAM_FMUNMAP)   == BAM_FMUNMAP)  * CRAM_M_UNMAP +
+	    //	((cr->flags & BAM_FMREVERSE) == BAM_FMREVERSE)* CRAM_M_REVERSE;
+	    //p->tlen = p->apos - cr->aend;
 
 	    // Clear detached from cr flags
-	    //cram_stats_del(c->stats[DS_CF], cr->cram_flags);
 	    cr->cram_flags &= ~CRAM_FLAG_DETACHED;
 	    cram_stats_add(c->stats[DS_CF], cr->cram_flags);
 
@@ -2672,6 +2911,7 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 	    hi->data.i = rnum;
 	    //HashTableDel(s->pair, hi, 0);
 	} else {
+	detached:
 	    //fprintf(stderr, "unpaired\n");
 
 	    /* Derive mate flags from this flag */
@@ -2691,6 +2931,7 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 
 	    cr->cram_flags |= CRAM_FLAG_DETACHED;
 	    cram_stats_add(c->stats[DS_CF], cr->cram_flags);
+	    cram_stats_add(c->stats[DS_NS], bam_mate_ref(b));
 	}
     }
 
@@ -2698,7 +2939,6 @@ static int process_one_read(cram_fd *fd, cram_container *c,
     cram_stats_add(c->stats[DS_MQ], cr->mqual);
 
     cr->mate_ref_id = bam_mate_ref(b);
-    cram_stats_add(c->stats[DS_NS], cr->mate_ref_id);
 
     if (!(bam_flag(b) & BAM_FUNMAP)) {
 	if (c->first_base > cr->apos)
@@ -2783,7 +3023,7 @@ int cram_put_bam_seq(cram_fd *fd, bam_seq_t *b) {
 
 	// Have we seen this reference before?
 	if (bam_ref(b) >= 0 && bam_ref(b) != curr_ref && !fd->embed_ref &&
-	    !fd->unsorted) {
+	    !fd->unsorted && multi_seq) {
 	    
 	    if (!c->refs_used) {
 		pthread_mutex_lock(&fd->ref_lock);
