@@ -94,23 +94,27 @@ static void dump_index(cram_fd *fd) {
 }
 #endif
 
-/*
- * Loads a CRAM .crai index into memory.
- *
- * Returns 0 for success
- *        -1 for failure
- */
-int cram_index_load(cram_fd *fd, char *fn) {
-    char line[1024], fn2[PATH_MAX];
-    zfp *fp;
-    cram_index *idx;
+typedef char * (*fgets_functions)(char * s, int size, void * fp);
+
+static char * zfgets_func(char * s, int size, void * fp)
+{
+    return zfgets(s,size,(zfp *)fp);
+}
+
+#if defined(CRAM_IO_CUSTOM_BUFFERING)
+static char * cram_io_input_buffer_fgets_func(char * s, int size, void * fp)
+{
+    return cram_io_input_buffer_fgets(s,size,fp);
+}
+#endif
+
+static int cram_index_load_private(cram_fd *fd, void * fp, fgets_functions fgets_func)
+{
+    char line[1024];
+    cram_index *idx = NULL;
     cram_index **idx_stack = NULL, *ep, e;
     int idx_stack_alloc = 0, idx_stack_ptr = 0;
 
-    /* Check if already loaded */
-    if (fd->index)
-	return 0;
-
     fd->index = calloc((fd->index_sz = 1), sizeof(*fd->index));
     if (!fd->index)
 	return -1;
@@ -123,22 +127,22 @@ int cram_index_load(cram_fd *fd, char *fn) {
     idx_stack = calloc(++idx_stack_alloc, sizeof(*idx_stack));
     idx_stack[idx_stack_ptr] = idx;
 
-    sprintf(fn2, "%s.crai", fn);
-    if (!(fp = zfopen(fn2, "r"))) {
-	perror(fn2);
-	free(idx_stack);
-	return -1; 
-    }
-
-    while (zfgets(line, 1024, fp)) {
+    while (fgets_func(line, 1024, fp)) {
 	/* 1.1 layout */
-	sscanf(line, "%d\t%d\t%d\t%"PRId64"\t%d\t%d",
-	       &e.refid,
-	       &e.start,
-	       &e.end,
-	       &e.offset,
-	       &e.slice,
-	       &e.len);
+	char *cp = line;
+        errno = 0;
+        e.refid  = strtol (cp, &cp, 10);
+        e.start  = strtol (cp, &cp, 10);
+        e.end    = strtol (cp, &cp, 10);
+        e.offset = strtoll(cp, &cp, 10);
+        e.slice  = strtol (cp, &cp, 10);
+        e.len    = strtol (cp, &cp, 10);
+
+        if (errno == EINVAL || errno == ERANGE) {
+            free(idx_stack);
+            return -1;
+        }
+
 	e.end += e.start-1;
 	//printf("%d/%d..%d\n", e.refid, e.start, e.end);
 
@@ -185,7 +189,6 @@ int cram_index_load(cram_fd *fd, char *fn) {
 	}
 	idx_stack[idx_stack_ptr] = idx;
     }
-    zfclose(fp);
     free(idx_stack);
 
     // dump_index(fd);
@@ -193,6 +196,87 @@ int cram_index_load(cram_fd *fd, char *fn) {
     return 0;
 }
 
+#if defined(CRAM_IO_CUSTOM_BUFFERING)
+/*
+ * Loads a CRAM .crai index into memory.
+ *
+ * Returns 0 for success
+ *        -1 for failure
+ */
+int cram_index_load_via_callbacks(
+    cram_fd *fd, char const * fn,
+    cram_io_allocate_read_input_t   callback_allocate_function,
+    cram_io_deallocate_read_input_t callback_deallocate_function        
+) {
+    cram_fd * input = NULL;
+    int r = -1;
+    static char const * indexsuffix = ".crai";
+    char * indexfn = NULL;
+    size_t const fnsize = strlen(fn);
+    size_t const suffixsize = strlen(indexsuffix);
+    size_t const indexfnsize = fnsize+suffixsize+1;
+    
+    if ( !(indexfn = (char *)malloc(indexfnsize)) ) {
+        r = -1;
+        goto cleanup;
+    }
+    
+    memcpy(indexfn,       fn,         fnsize);
+    memcpy(indexfn+fnsize,indexsuffix,suffixsize);
+    indexfn[fnsize+suffixsize] = 0;
+    
+    if ( ! (input = cram_io_open_by_callbacks(indexfn,callback_allocate_function,callback_deallocate_function,32*1024,1/* decompress */)) ) {
+        r = -1;
+        goto cleanup;
+    }
+
+    r = cram_index_load_private(fd,input,cram_io_input_buffer_fgets_func);
+    
+    cleanup:
+    if ( input ) {
+        cram_io_close(input,NULL);
+        input = NULL;
+    }
+    if ( indexfn ) {
+        free(indexfn);
+        indexfn = NULL;
+    }
+    
+    return r;
+}
+#endif
+
+/*
+ * Loads a CRAM .crai index into memory.
+ *
+ * Returns 0 for success
+ *        -1 for failure
+ */
+int cram_index_load(cram_fd *fd, char const *fn) {
+    zfp *fp = NULL;
+    char fn2[PATH_MAX];
+    int r = -1;
+    
+    /* Check if already loaded */
+    if (fd->index)
+	return 0;
+
+    /* copy filename */
+    sprintf(fn2, "%s.crai", fn);
+    
+    /* open index file */
+    if (!(fp = zfopen(fn2, "r"))) {
+	perror(fn2);
+	return -1; 
+    }
+    
+    r = cram_index_load_private(fd,fp,zfgets_func);
+
+    zfclose(fp);
+    
+    return r;
+}
+
 static void cram_index_free_recurse(cram_index *e) {
     if (e->e) {
 	int i;
@@ -290,7 +374,9 @@ cram_index *cram_index_query(cram_fd *fd, int refid, int pos,
 int cram_seek(cram_fd *fd, off_t offset, int whence) {
     char buf[65536];
 
-    if (fseeko(fd->fp, offset, whence) == 0)
+    fd->ooc = 0;
+
+    if (CRAM_IO_SEEK(fd, offset, whence) == 0)
 	return 0;
 
     if (!(whence == SEEK_CUR && offset >= 0))
@@ -299,7 +385,7 @@ int cram_seek(cram_fd *fd, off_t offset, int whence) {
     /* Couldn't fseek, but we're in SEEK_CUR mode so read instead */
     while (offset > 0) {
 	int len = MIN(65536, offset);
-	if (len != fread(buf, 1, len, fd->fp))
+	if (len != CRAM_IO_READ(buf, 1, len, fd))
 	    return -1;
 	offset -= len;
     }
@@ -337,6 +423,7 @@ int cram_seek_to_refpos(cram_fd *fd, cram_range *r) {
     if (fd->ctr) {
 	cram_free_container(fd->ctr);
 	fd->ctr = NULL;
+	fd->ooc = 0;
     }
 
     return 0;
@@ -408,17 +495,28 @@ int cram_index_build(cram_fd *fd, const char *fn_base) {
     off_t cpos, spos, hpos;
     zfp *fp;
     char fn_idx[PATH_MAX];
+    int seekable;
+    size_t len;
 
-    if (strlen(fn_base) > PATH_MAX-6)
+    if ((len=strlen(fn_base)) > PATH_MAX-6)
 	return -1;
 
-    sprintf(fn_idx, "%s.crai", fn_base);
+    if (len >= 5 && strcmp(&fn_base[len-5], ".crai") == 0)
+	strcpy(fn_idx, fn_base);
+    else
+	sprintf(fn_idx, "%s.crai", fn_base);
     if (!(fp = zfopen(fn_idx, "wz"))) {
         perror(fn_idx);
         return -1;
     }
 
-    cpos = ftello(fd->fp);
+    cpos = CRAM_IO_TELLO(fd);
+    if (cpos >= 0) {
+	seekable = 1;
+    } else {
+	seekable = 0;
+	cpos = fd->first_container;
+    }
     while ((c = cram_read_container(fd))) {
         int j;
 
@@ -427,7 +525,12 @@ int cram_index_build(cram_fd *fd, const char *fn_base) {
             return 1;
         }
 
-        hpos = ftello(fd->fp);
+	if (seekable) {
+	    hpos = CRAM_IO_TELLO(fd);
+	    assert(hpos == cpos + c->offset);
+	} else {
+	    hpos = cpos + c->offset;
+	}
 
         if (!(c->comp_hdr_block = cram_read_block(fd)))
             return 1;
@@ -443,15 +546,25 @@ int cram_index_build(cram_fd *fd, const char *fn_base) {
             cram_slice *s;
             int sz;
 
-            spos = ftello(fd->fp);
-            assert(spos - cpos - c->offset == c->landmark[j]);
+	    if (seekable) {
+		spos = CRAM_IO_TELLO(fd);
+		assert(spos - cpos - c->offset == c->landmark[j]);
+	    } else {
+		spos = cpos + c->offset + c->landmark[j];
+	    }
 
             if (!(s = cram_read_slice(fd))) {
 		zfclose(fp);
 		return -1;
 	    }
 
-            sz = (int)(ftello(fd->fp) - spos);
+	    if (seekable) {
+		sz = (int)(CRAM_IO_TELLO(fd) - spos);
+	    } else {
+		sz = j+1 < c->num_landmarks
+		    ? c->landmark[j+1] - c->landmark[j]
+		    : c->length - c->landmark[c->num_landmarks-1];
+	    }
 
 	    if (s->hdr->ref_seq_id == -2) {
 		cram_index_build_multiref(fd, c, s, fp,
@@ -466,9 +579,13 @@ int cram_index_build(cram_fd *fd, const char *fn_base) {
 
             cram_free_slice(s);
         }
-
-        cpos = ftello(fd->fp);
-        assert(cpos == hpos + c->length);
+	
+	if (seekable) {
+	    cpos = CRAM_IO_TELLO(fd);
+	    assert(cpos == hpos + c->length);
+	} else {
+	    cpos = hpos + c->length;
+	}
 
         cram_free_container(c);
     }
