@@ -57,6 +57,7 @@ extern "C" {
 
 #include <stdint.h>
 #include <io_lib/misc.h>
+#include <io_lib/bam.h>
 
 /**@{ ----------------------------------------------------------------------
  * ITF8 encoding and decoding.
@@ -220,21 +221,16 @@ char *cram_content_type2str(enum cram_content_type t);
 	(b)->data[(b)->byte++] = (c);	  \
     } while (0)
 
-/* Append via sprintf with 1 arg */
-#define BLOCK_APPENDF_1(b,buf,fmt, a1)			\
-    do {						\
-	int l = sprintf((buf), (fmt), (a1));		\
-	BLOCK_APPEND((b), (buf), l);			\
+/* Append a single unsigned integer */
+#define BLOCK_APPEND_UINT(b,i)                       \
+    do {                                             \
+	unsigned char *cp;                           \
+	BLOCK_GROW((b),11);                          \
+	cp = &(b)->data[(b)->byte];                  \
+	(b)->byte += append_uint(cp, (i)) - cp;	     \
     } while (0)
 
-/* Append via sprintf with 2 args */
-#define BLOCK_APPENDF_2(b,buf,fmt, a1,a2)		\
-    do {						\
-	int l = sprintf((buf), (fmt), (a1), (a2));	\
-	BLOCK_APPEND((b), (buf), l);			\
-    } while (0)
-
-#define BLOCK_UPLEN(b) \
+#define BLOCK_UPLEN(b)			\
     (b)->comp_size = (b)->uncomp_size = BLOCK_SIZE((b))
 
 /**@}*/
@@ -261,6 +257,8 @@ int cram_load_reference(cram_fd *fd, char *fn);
  */
 int refs2id(refs_t *r, SAM_hdr *bfd);
 
+refs_t *refs_load_fai(refs_t *r_orig, char *fn, int is_err);
+char *load_ref_portion(FILE *fp, ref_entry *e, int start, int end);
 void refs_free(refs_t *r);
 
 /*! Returns a portion of a reference sequence from start to end inclusive.
@@ -490,6 +488,35 @@ int cram_set_option(cram_fd *fd, enum cram_option opt, ...);
  */
 int cram_set_voption(cram_fd *fd, enum cram_option opt, va_list args);
 
+#if defined(CRAM_IO_CUSTOM_BUFFERING)
+/*
+ * Opens a CRAM file for input via callbacks
+ *
+ * Returns file handle on success
+ *         NULL on failure.
+ */
+extern cram_fd *cram_open_by_callbacks(
+    char const * filename,
+    cram_io_allocate_read_input_t   callback_allocate_function,
+    cram_io_deallocate_read_input_t callback_deallocate_function,
+    size_t const bufsize
+);
+
+extern cram_fd * cram_io_open(
+	char const * filename, 
+	char const * mode, 
+	char const * fmode
+);
+extern cram_fd * cram_io_open_by_callbacks(
+    char const * filename,
+    cram_io_allocate_read_input_t   callback_allocate_function,
+    cram_io_deallocate_read_input_t callback_deallocate_function,
+    size_t const bufsize,
+    int const decompress
+);
+extern cram_fd * cram_io_close(cram_fd * fd, int * fclose_result);
+#endif
+
 /**@}*/
 
 #ifdef __cplusplus
