@@ -111,16 +111,17 @@ enum cram_DS_ID {
     DS_aux    = 1, // aux_blk
     DS_aux_OQ = 2,
     DS_aux_BQ = 3,
-    DS_aux_BD = 4, // BD & BI
-    DS_aux_FZ = 5,
-    DS_aux_oq = 6, // other qualities
-    DS_aux_os = 7, // other sequences
-    DS_aux_oz = 8, // other strings
+    DS_aux_BD = 4,
+    DS_aux_BI = 5,
+    DS_aux_FZ = 6, // also ZM:B
+    DS_aux_oq = 7, // other qualities
+    DS_aux_os = 8, // other sequences
+    DS_aux_oz = 9, // other strings
     DS_ref,
     DS_RN, // name_blk
     DS_QS, // qual_blk
-    DS_SC, // soft_blk
     DS_IN, // base_blk
+    DS_SC, // soft_blk
 
     DS_BF, // start loop
     DS_CF,
@@ -144,13 +145,19 @@ enum cram_DS_ID {
     DS_RS,
     DS_PD,
     DS_HC,
+    DS_BB,
+    DS_QQ,
 
     DS_TN, // end loop
 
+    DS_RN_len,
+    DS_SC_len,
+    DS_BB_len,
+    DS_QQ_len,
+
     DS_TC, // CRAM v1.0 tags
     DS_TM, // test
     DS_TV, // test
-    DS_Qs, // BYTE_ARRAY version of QS
     
     DS_END,
 };
@@ -172,19 +179,20 @@ typedef struct {
 struct cram_slice;
 
 enum cram_block_method {
+    BM_ERROR  = -1,
     RAW    = 0,
     GZIP   = 1,
     BZIP2  = 2,
     LZMA   = 3,
-    ARITH0 = 4,
-    ARITH1 = 5,
-    RANS0  = 6,
-    RANS1  = 7,
-
-    GZIP_RLE = 8, // NB: not externalised in CRAM
+    RANS0  = 4,
+    RANS1  = 10,   // Not externalised; stored as RANS (generic)
+    GZIP_RLE = 11, // NB: not externalised in CRAM
+    ARITH0 = 16,   // unofficial & test only
+    ARITH1 = 17,
 };
 
 enum cram_content_type {
+    CT_ERROR           = -1,
     FILE_HEADER        = 0,
     COMPRESSION_HEADER = 1,
     MAPPED_SLICE       = 2,
@@ -445,6 +453,12 @@ typedef struct {
 	    int base;    // actual base & qual
 	    int qual;
 	} B;
+	struct {
+	    int pos;
+	    int code;
+	    int seq_idx; // index to s->seqs_blk
+	    int len;
+	} b;
 	struct {
 	    int pos;
 	    int code;
@@ -543,12 +557,13 @@ typedef struct cram_slice {
     cram_block *aux_OQ_blk;
     cram_block *aux_BQ_blk;
     cram_block *aux_BD_blk;
+    cram_block *aux_BI_blk;
     cram_block *aux_FZ_blk;
     cram_block *aux_oq_blk;
     cram_block *aux_os_blk;
     cram_block *aux_oz_blk;
 
-    HashTable *pair;         // for identifying read-pairs in this slice.
+    HashTable *pair[2];      // for identifying read-pairs in this slice.
 
     char *ref;               // slice of current reference
     int ref_start;           // start position of current reference;
@@ -631,8 +646,47 @@ typedef struct spare_bams {
     struct spare_bams *next;
 } spare_bams;
 
+#if defined(CRAM_IO_CUSTOM_BUFFERING)
+typedef size_t (*cram_io_C_FILE_fread_t)(void *ptr, size_t size, size_t nmemb, void *stream);
+typedef int    (*cram_io_C_FILE_fseek_t)(void * fd, off_t offset, int whence);
+typedef off_t  (*cram_io_C_FILE_ftell_t)(void * fd);
+
 typedef struct {
-    FILE          *fp;
+    void                   *user_data;
+    cram_io_C_FILE_fread_t  fread_callback;
+    cram_io_C_FILE_fseek_t  fseek_callback;
+    cram_io_C_FILE_ftell_t  ftell_callback;
+} cram_io_input_t;
+
+typedef cram_io_input_t * (*cram_io_allocate_read_input_t)(char const * filename, int const decompress);
+typedef cram_io_input_t * (*cram_io_deallocate_read_input_t)(cram_io_input_t * obj);
+
+typedef struct {
+    /* input buffer size */
+    size_t         fp_in_buf_size;
+    /* input buffer base pointer */
+    char          *fp_in_buffer;
+    /* position of buffer start in file */
+    uint64_t       fp_in_buf_start;
+    /* start of window pointer */
+    char          *fp_in_buf_pa;
+    /* window current pointer */
+    char          *fp_in_buf_pc;
+    /* window end pointer */
+    char          *fp_in_buf_pe;    
+} cram_fd_input_buffer;
+#endif
+
+typedef struct {
+    FILE                 *fp_in;
+    #if defined(CRAM_IO_CUSTOM_BUFFERING)
+    cram_fd_input_buffer *fp_in_buffer;
+    cram_io_input_t      *fp_in_callbacks;
+    cram_io_allocate_read_input_t   fp_in_callback_allocate_function;
+    cram_io_deallocate_read_input_t fp_in_callback_deallocate_function;
+    #endif
+    
+    FILE          *fp_out;
     int            mode;     // 'r' or 'w'
     int            version;
     cram_file_def *file_def;
@@ -674,7 +728,7 @@ typedef struct {
     int no_ref;
     int ignore_md5;
     int use_bz2;
-    int use_arith;
+    int use_rans;
     int use_lzma;
     int shared_ref;
     enum quality_binning binning;
@@ -709,6 +763,31 @@ typedef struct {
     int ooc;                            // out of containers.
 } cram_fd;
 
+#if defined(CRAM_IO_CUSTOM_BUFFERING)
+extern size_t cram_io_input_buffer_read(void *ptr, size_t size, size_t nmemb, cram_fd * fd);
+extern int cram_io_input_buffer_seek(cram_fd * fd, off_t offset, int whence);
+extern int cram_io_input_buffer_underflow(cram_fd * fd);
+extern char * cram_io_input_buffer_fgets(char * s, int size, cram_fd * fd);
+#endif
+
+#if defined(CRAM_IO_CUSTOM_BUFFERING)
+#define CRAM_IO_GETC(fd) ((fd->fp_in_buffer->fp_in_buf_pc!=fd->fp_in_buffer->fp_in_buf_pe) ? ((int)((unsigned char)(*(fd->fp_in_buffer->fp_in_buf_pc++)))) : cram_io_input_buffer_underflow(fd))
+#define CRAM_IO_READ(ptr, size, nmemb, fd) cram_io_input_buffer_read(ptr,size,nmemb,fd)
+#define CRAM_IO_SEEK(fd, offset, whence) cram_io_input_buffer_seek(fd, offset, whence)
+#define CRAM_IO_TELLO(fd) (fd->fp_in_buffer->fp_in_buf_start +(fd->fp_in_buffer->fp_in_buf_pc-fd->fp_in_buffer->fp_in_buf_pa))
+#define CRAM_IO_FGETS(s,size,fd) cram_io_input_buffer_fgets(s,size,fd)
+#else // ! CRAM_IO_CUSTOM_BUFFERING
+#define CRAM_IO_GETC(fd) getc(fd->fp_in)
+#define CRAM_IO_READ(ptr, size, nmemb, fd) fread(ptr,size,nmemb,fd->fp_in)
+#define CRAM_IO_TELLO(fd) ftello(fd->fp_in)
+#define CRAM_IO_SEEK(fd, offset, whence) fseeko(fd->fp_in,offset,whence)
+#define CRAM_IO_FGETS(s,size,fd) fgets(s,size,fd->fp_in)
+#endif // end CRAM_IO_CUSTOM_BUFFERING
+
+#define CRAM_IO_PUTC(c,fd) putc(c,fd->fp_out)
+#define CRAM_IO_WRITE(ptr, size, nmemb, fd) fwrite(ptr,size,nmemb,fd->fp_out)
+#define CRAM_IO_FLUSH(fd) (fd->fp_out ? fflush(fd->fp_out) : 0)
+
 // REQUIRED_FIELDS
 enum sam_fields {
     SAM_QNAME = 0x00000001,
@@ -723,6 +802,7 @@ enum sam_fields {
     SAM_SEQ   = 0x00000200,
     SAM_QUAL  = 0x00000400,
     SAM_AUX   = 0x00000800,
+    SAM_RGAUX = 0x00001000,
 };
 
 // Translation of required fields to cram data series
@@ -753,8 +833,12 @@ enum cram_fields {
     CRAM_PD = 0x00800000,
     CRAM_HC = 0x01000000,
     CRAM_SC = 0x02000000,
-    CRAM_aux= 0x04000000,
-    CRAM_ALL= 0x07ffffff,
+    CRAM_BB = 0x04000000,
+    CRAM_BB_len = 0x08000000,
+    CRAM_QQ = 0x10000000,
+    CRAM_QQ_len = 0x20000000,
+    CRAM_aux= 0x40000000,
+    CRAM_ALL= 0x7fffffff,
 };
 
 // A CIGAR opcode, but not necessarily the implications of it. Eg FC/FP may
@@ -764,8 +848,8 @@ enum cram_fields {
 #define CRAM_CIGAR (CRAM_FN | CRAM_FP | CRAM_FC | CRAM_DL | CRAM_IN | \
 		    CRAM_SC | CRAM_HC | CRAM_PD | CRAM_RS | CRAM_RL | CRAM_BF)
 
-#define CRAM_SEQ (CRAM_CIGAR | CRAM_BA | CRAM_QS | CRAM_BS | CRAM_BA | \
-		  CRAM_RL    | CRAM_AP)
+#define CRAM_SEQ (CRAM_CIGAR | CRAM_BA | CRAM_QS | CRAM_BS | \
+		  CRAM_RL    | CRAM_AP | CRAM_BB | CRAM_QQ)
 
 enum cram_option {
     CRAM_OPT_DECODE_MD,
@@ -788,6 +872,7 @@ enum cram_option {
     CRAM_OPT_USE_ARITH,
     CRAM_OPT_USE_LZMA,
     CRAM_OPT_REQUIRED_FIELDS,
+    CRAM_OPT_USE_RANS,
 };
 
 /* BF bitfields */
@@ -806,10 +891,11 @@ enum cram_option {
 #define DS_aux_OQ_S "\002"
 #define DS_aux_BQ_S "\003"
 #define DS_aux_BD_S "\004"
-#define DS_aux_FZ_S "\005"
-#define DS_aux_oq_S "\006"
-#define DS_aux_os_S "\007"
-#define DS_aux_oz_S "\010"
+#define DS_aux_BI_S "\005"
+#define DS_aux_FZ_S "\006"
+#define DS_aux_oq_S "\007"
+#define DS_aux_os_S "\010"
+#define DS_aux_oz_S "\011"
 
 #define CRAM_M_REVERSE  1
 #define CRAM_M_UNMAP    2
