@@ -30,6 +30,9 @@ extern "C" {
 /* Set by autoconf */
 #define SP_LITTLE_ENDIAN
 
+#define HAVE_STDIO_EXT_H
+#define CRAM_IO_CUSTOM_BUFFERING
+
 /* Mac FAT binaries or unknown. Auto detect based on CPU type */
 #if !defined(SP_BIG_ENDIAN) && !defined(SP_LITTLE_ENDIAN)
 
