@@ -741,6 +741,41 @@ int sam_hdr_update(SAM_hdr *hdr, SAM_hdr_type *type, ...) {
     return 0;
 }
 
+
+/*
+ * Returns the sort order:
+ */
+enum sam_sort_order sam_hdr_sort_order(SAM_hdr *hdr) {
+    return hdr->sort_order;
+}
+
+static enum sam_sort_order sam_hdr_parse_sort_order(SAM_hdr *hdr) {
+    HashItem *hi;
+    enum sam_sort_order so;
+
+    so = ORDER_UNKNOWN;
+    if ((hi = HashTableSearch(hdr->h, "HD", 2))) {
+	SAM_hdr_type *ty = hi->data.p;
+	SAM_hdr_tag *tag;
+        for (tag = ty->tag; tag; tag = tag->next) {
+	    if (tag->str[0] == 'S' && tag->str[1] == 'O') {
+		if (strcmp(tag->str+3, "unsorted") == 0)
+		    so = ORDER_UNSORTED;
+		else if (strcmp(tag->str+3, "queryname") == 0)
+		    so = ORDER_NAME;
+		else if (strcmp(tag->str+3, "coordinate") == 0)
+		    so = ORDER_COORD;
+		else
+		    fprintf(stderr, "Unknown sort order field: %s\n",
+			    tag->str+3);
+	    }
+	}
+    }
+
+    return so;
+}
+
+
 /*
  * Reconstructs the dstring from the header hash table.
  * Returns 0 on success
@@ -901,6 +936,9 @@ SAM_hdr *sam_hdr_parse(const char *hdr, int len) {
 	return NULL;
     }
 
+    /* Obtain sort order */
+    sh->sort_order = sam_hdr_parse_sort_order(sh);
+
     //sam_hdr_dump(sh);
     //sam_hdr_add(sh, "RG", "ID", "foo", "SM", "bar", NULL);
     //sam_hdr_rebuild(sh);
