@@ -431,6 +431,7 @@ t_pool *t_pool_init(int qsize, int tsize) {
     p->nwaiting = 0;
     p->shutdown = 0;
     p->head = p->tail = NULL;
+    p->t_stack = NULL;
 #ifdef DEBUG_TIME
     p->total_time = p->wait_time = 0;
 #endif
@@ -563,7 +564,6 @@ int t_pool_dispatch2(t_pool *p, t_results_queue *q,
     if (p->njobs >= p->qsize && nonblock == 1) {
 	pthread_mutex_unlock(&p->pool_m);
 	errno = EAGAIN;
-	free(j);
 	return -1;
     }
 
@@ -725,6 +725,9 @@ void t_pool_destroy(t_pool *p, int kill) {
 		p->t[i].wait_time / 1000000.0);
 #endif
 
+    if (p->t_stack)
+	free(p->t_stack);
+
     free(p->t);
     free(p);
 
